"""
Storage retrieval for Chalice
"""
from typing import Union, Optional

# from chalice.app import Request, Response
# from genericsuite.util.blueprint_one import BlueprintOne
from genericsuite.util.framework_abs_layer import Request, Response, BlueprintOne
from genericsuite.util.app_logger import log_debug

from genericsuite.util.aws import storage_retieval
from genericsuite.util.utilities import (
    return_resultset_jsonified_or_exception,
)

DEBUG = True
bp = BlueprintOne(__name__)


def debug_args(args1, kwargs1):
    """ Debug args """
    if DEBUG:
        log_debug(f"args1: {args1}")
        log_debug(f"kwargs1: {kwargs1}")


@bp.route('/', methods=['GET'])
def storage_retrieval_no_item_id_endpoint(
    *args,
    **kwargs,
) -> Response:
    """ Get file from ecrypted URL """
    debug_args(args, kwargs)
    request = bp.get_current_request()  #   bp.current_app.current_request
    # Report the error ASR-E1010
    item_id = None
    other_params = kwargs.get('other_params', {}) or {}
    other_params['response_type'] = other_params.get('response_type') or "streaming"
    return storage_retieval_chalice(request=request, blueprint=bp,
        item_id=item_id, other_params=other_params)


@bp.route('/{item_id}', methods=['GET'])
def storage_retrieval_endpoint(
    # item_id: str,
    *args,
    **kwargs,
) -> Response:
    """ Get file from ecrypted URL """
    debug_args(args, kwargs)
    request = bp.get_current_request()  #   bp.current_app.current_request
    item_id = kwargs.get('item_id')
    other_params = kwargs.get('other_params', {}) or {}
    other_params['response_type'] = other_params.get('response_type') or "streaming"
    return storage_retieval_chalice(request=request, blueprint=bp,
        item_id=item_id, other_params=other_params)


@bp.route('/{item_id}/{response_type}', methods=['GET'])
def storage_retrieval_with_response_type_endpoint(
    # other_params: Optional[Union[dict, None]] = None,
    # item_id: str,
    # response_type: str,
    *args, **kwargs,
) -> Response:
    """ Get file from ecrypted URL """
    request = bp.get_current_request()  #   bp.current_app.current_request
    item_id = kwargs.get('item_id')
    other_params = kwargs.get('other_params', {}) or {}
    other_params['response_type'] = kwargs.get('response_type') or "streaming"
    return storage_retieval_chalice(request=request, blueprint=bp,
        item_id=item_id, other_params=other_params)


def storage_retieval_chalice(
    request: Request,
    blueprint: BlueprintOne,
    item_id: str,
    other_params: Optional[Union[dict, None]] = None,
) -> Response:
    """
    Get S3 bucket content from encrypted item_id
    Args:
        request (Request): The request object.
        other_params (dict, optional): Other parameters. Defaults to None.
        item_id (str, optional): The item_id with the encrypted elements
            (bucket_name, separator and key). Defaults to None.
    Returns:
        Union[Response, StreamingResponse]: The object as streaming response
            or error response.
    """
    # resultset = storage_retieval(request, item_id, other_params)
    resultset = storage_retieval(request=request, blueprint=blueprint,
        item_id=item_id, other_params=other_params)
    if resultset.get('error'):
        return return_resultset_jsonified_or_exception(
            resultset
        )
    if other_params.get('response_type') == "streaming":
        # Return the file content as a Streaming Response
        return Response(
            body=resultset['content'],
            status_code=200,
            headers={'Content-Type': 'application/octet-stream'}
        )
    # Return the file content as a normal Response
    headers = {
        'Content-Type': resultset['mime_type'],
        'Content-Disposition': 'attachment; filename=' \
            f'"{resultset["filename"]}"',
    }
    return Response(
        body=resultset['content'],
        status_code=200,
        headers=headers
    )



"""
Secrets config management
"""
from typing import Callable
import os

from genericsuite.util.aws_secrets import (
    get_cache_secret as get_aws_secrets,
    get_secrets_cache_filename as get_aws_cache_filename,
)
from genericsuite.util.gcp_secrets import (
    get_cache_secret as get_gcp_secrets,
    get_secrets_cache_filename as get_gcp_cache_filename,
)
from genericsuite.util.azure_secrets import (
    get_cache_secret as get_azure_secrets,
    get_secrets_cache_filename as get_azure_cache_filename,
)

DEBUG = False


def get_secrets_from_iaas(get_default_resultset: Callable, logger: Callable
                          ) -> dict:
    """
    Get secrets from the iaas (AWS, GCP, Azure) and set environment variables.

    If the GET_SECRETS_ENABLED environment variable is "1" (default value),
    there must be environment variables set for:
        CLOUD_PROVIDER, APP_NAME, APP_STAGE,
        AWS_REGION or GCP_REGION or AZURE_REGION
    and the mandatory environment variables must be defined in the cloud
    provider secrets manager.
    Check the script "scripts/aws_secrets/aws_secrets_manager.sh" in the
    GenericSuite Backend Scripts package for more information.

    If GET_SECRETS_ENABLED is not "1", the mandatory environment variables
    must be set:
        APP_DB_URI, APP_SUPERADMIN_EMAIL, APP_DB_NAME, APP_DB_ENGINE,
        APP_NAME, APP_SECRET_KEY, APP_HOST_NAME, STORAGE_URL_SEED,
        GIT_SUBMODULE_LOCAL_PATH
    """
    result = get_default_resultset()
    if os.environ.get("GET_SECRETS_ENABLED", "1") != "1":
        _ = DEBUG and logger.debug("GET_SECRETS_ENABLED set to 0..." +
                                   " getting all envvars from environment")
        return result
    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    if not cloud_provider:
        result["error"] = True
        result["error_message"] = "ERROR: CLOUD_PROVIDER not set [GSFI-E010]"
        return result
    if cloud_provider.upper() == "AWS":
        iaas_secrets = get_aws_secrets(get_default_resultset, logger)
    elif cloud_provider.upper() == "GCP":
        iaas_secrets = get_gcp_secrets(get_default_resultset, logger)
    elif cloud_provider.upper() == "AZURE":
        iaas_secrets = get_azure_secrets(get_default_resultset, logger)
    else:
        result["error"] = True
        result["error_message"] = \
            "ERROR: CLOUD_PROVIDER not supported [GSFI-E020]"
        return result
    if iaas_secrets["error"]:
        return iaas_secrets
    for key, value in iaas_secrets["resultset"].items():
        _ = DEBUG and logger.debug(f"EnvVar {key} = {value}")
        os.environ[key] = value
    return result


def get_secrets_cache_filename(secret_type: str = "") -> str:
    """
    Get secrets cache filename
    """
    cloud_provider = os.environ.get("CLOUD_PROVIDER")
    if not cloud_provider:
        error_message = "ERROR: CLOUD_PROVIDER not set [GSCF-E010]"
        raise Exception(error_message)
    filename = None
    if cloud_provider.upper() == "AWS":
        filename = get_aws_cache_filename(secret_type)
    elif cloud_provider.upper() == "GCP":
        filename = get_gcp_cache_filename(secret_type)
    elif cloud_provider.upper() == "AZURE":
        filename = get_azure_cache_filename(secret_type)
    else:
        error_message = "ERROR: CLOUD_PROVIDER not supported [GSCF-E020]"
        raise Exception(error_message)
    return filename



"""
Security module
"""
from genericsuite.util.app_context import AppContext
from genericsuite.util.app_logger import log_debug
from genericsuite.util.generic_endpoint_helpers import GenericEndpointHelper
from genericsuite.util.utilities import (
    get_default_resultset, get_query_params, get_request_body
)

DEBUG = False


def get_user_groups(user_data: dict) -> list:
    """Get the user groups."""
    user_groups = []
    if "groups" in user_data:
        user_groups = user_data["groups"]
    if "superuser" in user_data and user_data["superuser"] == '1':
        user_groups.append('admin')
    user_groups.append('users')
    return user_groups


def get_user_authorized_menu(menu_data: list, user_data: dict) -> dict:
    """
    Get the authorized menu options.
    """
    menu_response = get_default_resultset()
    user_groups = get_user_groups(user_data)
    if DEBUG:
        log_debug(f'GAMO-1) user_data: {user_data}')
        log_debug(f'GAMO-2) user_groups: {user_groups}')
        # log_debug(f'GAMO-3) menu_data: {menu_data}')
    authorized_menu_options = authorize_menu_options(
        menu_data, user_groups
    )
    # if DEBUG:
    #     log_debug(f'GAMO-3) authorized_menu_options: {authorized_menu_options}')
    menu_response["resultset"] = authorized_menu_options
    return menu_response


def authorize_menu_options(menu_data: list, user_groups: list) -> list:
    """
    Authorize menu options based on user groups.

    Args:
        menu_data (list): The menu data to be authorized.
        user_groups (list): The groups of the current user.

    Returns:
        list: The authorized menu options.
    """
    authorized_menu_options = []
    for menu_option in menu_data:
        if "sub_menu_options" in menu_option:
            menu_option["sub_menu_options"] = \
                authorize_menu_options(
                    menu_option["sub_menu_options"], user_groups
                )
        if "sec_group" in menu_option:
            if menu_option["sec_group"] in user_groups:
                authorized_menu_options.append(menu_option)
        else:
            authorized_menu_options.append(menu_option)
    return authorized_menu_options


def get_authorized_menu_options(app_context: AppContext,
                                url_prefix: str = __name__
                                ) -> (dict, dict, list):
    """
    Get the authorized menu options.

    Args:
        app_context (AppContext): The application conntext
        url_prefix (str): The URL prefix.

    Returns:
        (dict, dict, list): A tuple containing the authorized menu options),
        the current user data, and the user groups list.
    """
    ep_helper = GenericEndpointHelper(
        app_context=app_context,
        json_file="app_main_menu",
        url_prefix=url_prefix,
        db_type="menu_options"
    )
    menu_response = get_default_resultset()
    user_groups = []
    user_data = app_context.get_user_data()
    if not user_data:
        menu_response = app_context.get_error_resultset()
        return menu_response, user_data, user_groups
    menu_data = ep_helper.generic_raw_json()
    if menu_data['error']:
        menu_response["error"] = True
        menu_response["error_message"] = menu_data['error_message']
        return menu_response, user_data, user_groups
    menu_response = get_user_authorized_menu(menu_data["resultset"],
                                             user_data)
    user_groups = get_user_groups(user_data)
    return menu_response, user_data, user_groups


def get_element_option(menu_data: list, element: str) -> dict:
    """
    Get the element option.
    """
    element_option = None
    for menu_option in menu_data:
        if "sub_menu_options" in menu_option:
            element_option = get_element_option(
                menu_option["sub_menu_options"], element
            )
            if element_option:
                if DEBUG:
                    log_debug(f'GEO-2) element: {element} Found!' +
                              f' -> {element_option}')
                return element_option
        if "element" in menu_option:
            if DEBUG:
                log_debug(f'GEO-1) element: {element} | ' +
                          f'menu_option.element: {menu_option["element"]}')
            if menu_option["element"] == element:
                element_option = menu_option
                if DEBUG:
                    log_debug(f'GEO-3) element: {element} Found!' +
                              f' -> {element_option}')
                return element_option
    return element_option


def get_option_access(app_context: AppContext, element: str) -> dict:
    """
    Get the user data from the app_context. Then, get the user groups from the
    user data, calling get_user_groups().
    Finally, call authorize_menu_options() with the menu data and user groups.
    Then verify in the authorized menu options if the option element
    is present.
    """
    oa_response = get_default_resultset()
    menu_response, _, user_groups = get_authorized_menu_options(app_context)
    if menu_response['error']:
        return menu_response
    menu_data = menu_response['resultset']
    element_option = get_element_option(menu_data, element)
    if not element_option:
        oa_response['error'] = True
        oa_response['error_message'] = \
            f'Element {element} not found [SOAV-010]'
    else:
        if 'sec_group' in element_option and \
           element_option['sec_group'] not in user_groups:
            oa_response['error'] = True
            oa_response['error_message'] = 'Access Denied [SOAV-020]'
    return oa_response


def verify_user_filter(app_context: AppContext) -> dict:
    """
    Verify the user filter in the endpoints that requires the user_id
    parameter.
    """
    request = app_context.get_request()
    vuf_response = get_default_resultset()
    if request.method.upper() in ['GET', 'DELETE']:
        query_params = get_query_params(request)
        user_id = query_params.get('user_id')
    else:  # request.method.upper() in ['POST', 'PUT']:
        form_data = get_request_body(request)
        user_id = form_data.get('user_id')
    if DEBUG:
        log_debug(f'VUF-1) VERIFY_USER_FILTER | user_id: {user_id}')
    if not user_id:
        vuf_response['error'] = True
        # Elements to build the Response()
        vuf_response['response'] = {
            "body": "user_id is required",
            "status_code": 400,
            "headers": {'Content-Type': 'text/plain'},
        }
    return vuf_response



"""
System users operations (CRUD, login, database test, super-admin creation)
"""
from typing import Optional

# from chalice.app import Request, Response
# from genericsuite.util.blueprint_one import BlueprintOne
from genericsuite.util.framework_abs_layer import (
    Request,
    Response,
    BlueprintOne
)

from genericsuite.util.jwt import (
    request_authentication,
    AuthorizedRequest,
)

from genericsuite.models.users.users import (
    # users_crud as users_crud_model,
    test_connection_handler as test_connection_handler_model,
    login_user as login_user_model,
    super_admin_create as super_admin_create_model,
    get_current_user_data,
)

bp = BlueprintOne(__name__)


HEADER_CREDS_ENTRY_NAME = 'Authorization'
DEBUG = False


# @bp.route(
#     '/',
#     methods=['GET', 'POST', 'PUT', 'DELETE'],
#     authorizor=request_authentication(),
# )
# def users_crud(request: AuthorizedRequest,
#     other_params: Optional[dict] = None) -> Response:
#     """ User's CRUD operations (create, read, update, delete) """
#     return users_crud_model(request, other_params)


@bp.route(
    '/test',
    authorizor=request_authentication(),
)
def test_connection_handler(
    request: Request,
    other_params: Optional[dict] = None
) -> Response:
    """Connection handler test"""
    return test_connection_handler_model(request, other_params)


@bp.route(
    '/login',
    methods=['GET', 'POST']
)
def login_user(
    request: Request,
    other_params: Optional[dict] = None
) -> Response:
    """User login"""
    # return login_user_model(request, other_params)
    return login_user_model(
        request=request, blueprint=bp,
        other_params=other_params)


@bp.route(
    '/supad-create',
    methods=['POST']
)
def super_admin_create(
    request: Request,
    other_params: Optional[dict] = None
) -> Response:
    """Super admin user emergency creation"""
    # return super_admin_create_model(request, other_params)
    return super_admin_create_model(
        request=request, blueprint=bp,
        other_params=other_params)


@bp.route(
    '/current_user_d',
    methods=['GET'],
    authorizor=request_authentication(),
)
def current_user_d(
    request: AuthorizedRequest,
    other_params: Optional[dict] = None
) -> Response:
    """
    Get current authenticated user data
    """
    return get_current_user_data(request, bp, other_params)




"""
Flask abstraction layer
"""
import os
import importlib

# https://stackoverflow.com/questions/47277374/flask-get-current-blueprint-webroot
from flask import Blueprint as FlaskBlueprint


DEBUG = False
FRAMEWORK_LOADED = False
FRAMEWORK = os.environ.get('CURRENT_FRAMEWORK', '').lower()
if FRAMEWORK == 'flask':
    try:
        framework_module = importlib.import_module(FRAMEWORK)
        FRAMEWORK_LOADED = True
        if DEBUG:
            print(f'Flask abstraction | framework_module: {framework_module}')


        class FrameworkClass(framework_module.Flask):
            """
            Framkework class cloned from the selected framework super class.
            """


        class Request(framework_module.request):
            """
            Request class cloned from the selected Request framework super class.
            This class is the one to be imported by the project modules
            """


        class Response(framework_module.response):
            """
            Response class cloned from the selected Response framework super class.
            This class is the one to be imported by the project modules
            """


        # class Blueprint(framework_module.Blueprint):
        class Blueprint(FlaskBlueprint):
            """
            Blueprint class cloned from the selected Blueprint framework super class.
            This class is the one to be imported by the project modules
            """


        class BlueprintOne(Blueprint):
            """
            Class to register a new route with optional schema validation and authorization.
            """

    except ImportError as err:
        raise ImportError(f"Unable to import '{FRAMEWORK}': {err}") from None



"""
AWS Utilities
"""
from typing import Optional, Union, Any
from datetime import datetime
import os
import json
from urllib.parse import urlparse

import requests

import boto3
from botocore.exceptions import NoCredentialsError

from genericsuite.config.config_from_db import app_context_and_set_env
from genericsuite.util.framework_abs_layer import Request
from genericsuite.util.app_logger import log_debug, log_error
from genericsuite.util.utilities import (
    get_default_resultset,
    error_resultset,
    get_mime_type,
    get_file_extension,
)
from genericsuite.util.encryption import decrypt_string, encrypt_string
from genericsuite.util.file_utilities import temp_filename
from genericsuite.config.config import Config

DEBUG = True

STORAGE_URL_SEPARATOR = '||'
STORAGE_ENCRYPTION = os.environ.get('STORAGE_ENCRYPTION', '') == '1'


def s3_base_url(bucket_name: str) -> str:
    """
    Returns the S3 base URL.

    Args:
        bucket_name (str): The base path of the S3 bucket.

    Returns:
        str: The S3 base URL.
    """
    return f"https://{bucket_name}.s3.amazonaws.com"


def upload_nodup_file_to_s3(
    file_path: str,
    original_filename: str,
    bucket_name: str,
    sub_dir: Optional[Union[str, None]] = None,
    public_file: bool = False,
) -> dict:
    """
    Uploads a local file to an S3 bucket.

    Args:
        file_path (str): The local path of the file.
        original_filename (str): If the original filename is specified,
        uses it, if not use the local path's file name.
        bucket_name (str): The base path of the S3 bucket.
        sub_dir (str): intermediate path. Defaults to None.
        public_file (bool): True to make the file public.

    Returns:
        str: The S3 URL of the uploaded file.
        str: the final filename (with a date/time prefix)
        str: The eventual error message
    """

    # If the original filename is specified, uses it, if not
    # use the local path's file name
    if original_filename:
        final_filename = original_filename
    else:
        final_filename = os.path.basename(file_path)

    # Add date/time to avoid file name duplicates
    final_filename = s3_nodup_filename(final_filename)

    # Construct the S3 bucket path
    dest_path = (f"{sub_dir}/" if sub_dir else "") + final_filename

    return upload_file_to_s3(
        bucket_name=bucket_name,
        source_path=file_path,
        dest_path=dest_path,
        public_file=public_file,
    )


def upload_file_to_s3(
    bucket_name: str,
    source_path: str,
    dest_path: str,
    public_file: bool = False
) -> dict:
    """
    Uploads a local file to an S3 bucket.

    Args:
        bucket_name (str): The base path of the S3 bucket.
        source_path (str): The local path of the file.
        dest_path (str): The S3 path of the file.
        public_file (bool): True to make the file public (ACL public-read)

    Returns:
        dict: a Dict with the following elements:
            public_url (str): The S3 (or encrypted) URL of the uploaded file.
            final_filename (str): the final filename (with a date/time prefix)
            error (bool): True if there was any error.
            error_message (str): The eventual error message
    """
    error = None
    result = get_default_resultset()

    # Initialize S3 client
    s3_client = boto3.client('s3')

    try:
        # Upload the file
        s3_client.upload_file(
            source_path,
            bucket_name,
            dest_path,
        )
        log_debug(f"\nFile uploaded successfully S3 Bucket: {bucket_name}" +
                  f" | path: {dest_path}")
    except FileNotFoundError:
        error = f"The file {source_path} was not found."
        log_debug(error)
    except NoCredentialsError:
        error = "Credentials not available for S3 upload."
        log_debug(error)

    if public_file and not error:
        try:
            # Enable public access for the uploaded file
            # It seems that setting the ACL to 'public-read'
            # is not turning the object public...
            # As an alternative, we can use the put_bucket_policy method
            # to make the bucket public
            policy = {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Sid": "AddPerm",
                        "Effect": "Allow",
                        "Principal": "*",
                        "Action": "s3:GetObject",
                        "Resource": f"arn:aws:s3:::{bucket_name}/{dest_path}"
                    }
                ]
            }
            s3_client.put_bucket_policy(
                Bucket=bucket_name,
                Policy=json.dumps(policy)
            )
            log_debug(f"Bucket policy for '{dest_path}' set to public-read" +
                      " successfully")
        # except s3_client.exceptions.S3Error as e:
        except Exception as err:
            error = f"Failed to set ACL for {dest_path}: {err}"
            log_debug(error)

    if STORAGE_ENCRYPTION:
        # Return the encrypted S3 URL of the uploaded file
        public_url = get_storage_masked_url(bucket_name, dest_path)
    else:
        # Return the public S3 URL of the uploaded file
        public_url = f"{s3_base_url(bucket_name)}/{dest_path}"

    # Final filanme is the file name in the S3 destination path.
    final_filename = os.path.basename(dest_path)

    log_debug(f"Public url: {public_url}")
    log_debug(f"Final filename: {final_filename}")
    log_debug(f"Error: {error}\n")

    result['public_url'] = public_url
    result['final_filename'] = final_filename
    result['error'] = error is not None
    result['error_message'] = error
    return result
    # return public_url, final_filename, error


def s3_nodup_filename(file_name):
    """
    Generate a unique filename for S3 to prevent overwriting existing files.

    Appends the current date and time to the original file name, ensuring that
    the uploaded file does not duplicate an existing file in the S3 bucket.
    Also replaces blanks with "_".

    Parameters:
        file_name (str): The original file name.

    Returns:
        str: A unique S3 filename with the current date and time appended.
    """
    result = f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}" + \
             f"_{file_name}"
    result = result.replace(" ", "_")
    result = result.replace(" ", "_")
    return result


def save_file_from_url(
    url: str,
    bucket_name: str,
    sub_dir: Optional[str] = None,
    original_filename: Optional[str] = None
) -> dict:
    """
    Save an image from a URL to AWS S3

    Args:
        url (str): The URL to get the image.
        bucket_name (str): The base path of the S3 bucket.
        sub_dir (str): intermediate path. Defaults to None.
        original_filename (str): original file name. Defaults to None.
        If original_filename is None, it will be the last element
        of the URL split by "/".

    Returns:
        (str, str, int, str): a tuple with the following elements:
        public_url (str): URL for the image.
        final_filename (str): file name of the image with date/time added.
        file_size (int): the file size in bytes.
        error (str): the eventual error message or None if no errors
    """
    settings = Config()
    result = get_default_resultset()
    if not original_filename:
        original_filename = url.split('/')[-1]
    response = requests.get(url, timeout=10)    # 10 seconds timeout
    tmp_file_path = settings.TEMP_DIR + '/' + original_filename
    with open(tmp_file_path, 'wb') as f_handler:
        f_handler.write(response.content)
    # file_size = os.stat(tmp_file_path).st_size
    result['file_size'] = os.stat(tmp_file_path).st_size
    # public_url, final_filename, error = upload_nodup_file_to_s3(
    upload_result = upload_nodup_file_to_s3(
        file_path=tmp_file_path,
        original_filename=original_filename,
        bucket_name=bucket_name,
        sub_dir=sub_dir,
    )
    # attachment_url = public_url
    os.remove(tmp_file_path)  # Clean up the temporary file

    result['public_url'] = upload_result['public_url']
    result['final_filename'] = upload_result['final_filename']
    result['error'] = upload_result['error']
    result['error_message'] = upload_result['error_message']
    return result
    # return attachment_url, final_filename, file_size, error


def remove_from_s3(bucket_name: str, key: str) -> dict:
    """
    Remove an object from an S3 bucket.

    Args:
        bucket_name (str): The base path of the S3 bucket.
        key (str): The S3 key of the object to be removed.
    """
    result = get_default_resultset()
    try:
        s3 = boto3.client('s3')
        s3.delete_object(Bucket=bucket_name, Key=key)
        log_debug(f"Object removed from S3: {bucket_name}/{key}")
    except Exception as err:
        result['error'] = True
        result['error_message'] = f"Failed to remove object from S3: {err}"
        log_debug(result['error_message'])
    return result


def get_s3_object(bucket_name: str, key: str) -> dict:
    """
    Get an object from an S3 bucket.

    Args:
        bucket_name (str): The base path of the S3 bucket.
        key (str): The S3 key of the object to be retrieved.

    Returns:
        dict: The object as a standard resultset dictionary,
            with the file content in the 'content' element
            or error/error_message elements.
    """
    result = get_default_resultset()
    try:
        s3 = boto3.client('s3')
        obj = s3.get_object(Bucket=bucket_name, Key=key)
        # result['content'] = obj['Body'].read().decode('utf-8')
        result['content'] = obj['Body'].read()
        log_debug(f"Object retrieved from S3: {bucket_name}/{key}")
    except Exception as err:
        result['error'] = True
        result['error_message'] = f"Failed to retrieve object from S3: {err}"
        log_error(result['error_message'])
    return result


def download_s3_object(bucket_name: str, key: str,
                       local_file_path: Optional[str] = None) -> dict:
    """
    Download an object from an S3 bucket in a temporary path.

    Args:
        bucket_name (str): The base path of the S3 bucket.
        key (str): The S3 key of the object to be retrieved.
        local_file_path (str, optional): The temporary path to download
            the file.
            If None, it will use a random-generated temporary path.
            Defaults to None.

    Returns:
        dict: The object as a standard resultset dictionary,
            with the file temp_path/filename in the 'local_file_path' element
            or error/error_message elements.
    """
    result = get_default_resultset()
    if not local_file_path:
        local_file_path = temp_filename(get_file_extension(file_path=key))
    try:
        s3_client = boto3.client('s3')
        s3_client.download_file(bucket_name, key, local_file_path)
        result['local_file_path'] = local_file_path
        log_debug(f"Object downloaded from S3: {bucket_name}/{key}")
    except Exception as err:
        result['error'] = True
        result['error_message'] = \
            f"ERROR-DS3O-010 - Failed to download object: {err}"
        log_error(result['error_message'])
    return result


def get_storage_masked_url(
    bucket_name: str, key: str,
    hostname: Optional[Union[str, None]] = None
):
    """
    Get S3 bucket masked URL
    Args:
        bucket_name (str): The base path of the S3 bucket.
        key (str): The S3 key of the object to be retrieved.
    Returns:
        str: The S3 bucket masked URL
    """
    settings = Config()
    if hostname:
        # If hostname is provided, it's a development environment.
        # Use it instead of the default hostname
        protocol = os.environ.get('URL_MASK_EXTERNAL_PROTOCOL')
        protocol = protocol or os.environ.get('RUN_PROTOCOL', 'https')
    else:
        hostname = settings.APP_HOST_NAME
        protocol = 'https'
    extension = key.split('.')[-1] if '.' in key else ''
    extension = '.' + extension if extension else ''
    key = key.rsplit('.', 1)[0] if '.' in key else key
    return protocol + '://' + hostname + "/asset/" + \
        encrypt_string(
            settings.STORAGE_URL_SEED,
            bucket_name + STORAGE_URL_SEPARATOR + key
        ) + \
        extension


def storage_retieval(
    request: Request,
    blueprint: Any,
    item_id: Union[str, None],
    other_params: Optional[dict] = None,
) -> dict:
    """
    Get S3 bucket content from encrypted item_id
    Args:
        request (Request): the request object.
        blueprint (Any): the blueprint object,
        other_params (dict, optional): Other parameters. Defaults to None.
        item_id (str, optional): The item_id with the encrypted elements
            (bucket_name, separator and key). Defaults to None.
    Returns:
        dict: The object as a standard resultset dictionary,
            with the file content in the 'content' element,
            the mime type in the 'mime_type' element,
            the file name in the 'filename' element (S3 key),
            the downloaded local file path in 'local_file_path' element,
            or error/error_message elements.
    """
    if other_params is None:
        other_params = {}
    if not other_params.get('mode'):
        # Default mode is 'get', the other option is 'download'
        other_params['mode'] = 'get'
    # Set environment variables from the database configurations.
    app_context = app_context_and_set_env(request=request, blueprint=blueprint)
    if app_context.has_error():
        return app_context.get_error_resultset()
    settings = Config(app_context)
    if not item_id:
        return error_resultset("Item ID is required", "ASR-E1010")
    # Decrypt item_id to get the bucket_name and key (filespec)
    extension = item_id.split('.')[-1] if '.' in item_id else ''
    extension = '.' + extension if extension else ''
    raw_item_id = item_id.rsplit('.', 1)[0] if '.' in item_id else item_id
    decripted_item_id = decrypt_string(settings.STORAGE_URL_SEED, raw_item_id)
    if not decripted_item_id:
        return error_resultset("Invalid asset", "ASR-E1020")
    # Get bucket name and file path
    bucket_name, key = decripted_item_id.split(STORAGE_URL_SEPARATOR)
    key = key+extension
    # Get the file content
    if DEBUG:
        log_debug(f">> bucket_name: {bucket_name} | key: {key}")
    if other_params['mode'] == 'get':
        retrieval_resultset = get_s3_object(bucket_name=bucket_name, key=key)
    else:
        retrieval_resultset = download_s3_object(bucket_name=bucket_name,
                                                 key=key)
    if retrieval_resultset.get('error'):
        # return retrieval_resultset
        return error_resultset(retrieval_resultset['error_message'],
                               "ASR-E1030")
    retrieval_resultset['mime_type'] = get_mime_type(key)
    retrieval_resultset['filename'] = key
    return retrieval_resultset


def prepare_asset_url(public_url):
    """
    Prepares the asset (image, sound, file) URL for the GPT4 vision message
    content. If DEV_MASK_EXT_HOSTNAME is set, it will be prepended to the URL.
    Args:
        public_url (str): The public URL of the image.
    Returns:
        str: The prepared image URL.
    """
    dev_mask_ext_hostname = os.environ.get('DEV_MASK_EXT_HOSTNAME')
    final_public_url = public_url
    if dev_mask_ext_hostname and STORAGE_ENCRYPTION:
        parsed_url = urlparse(public_url)
        final_public_url = dev_mask_ext_hostname + parsed_url.path
        if DEBUG:
            log_debug("prepare_asset_url" +
                      f" | dev_mask_ext_hostname: {dev_mask_ext_hostname}"
                      f" | parsed_url: {parsed_url}"
                      f" | final_public_url: {final_public_url}")
    return final_public_url



"""
App main module (create_app) for FastAPI
"""
from typing import Any

# import importlib
# import logging

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from mangum import Mangum

from genericsuite.util.app_logger import log_info
# from genericsuite.util.app_logger import log_debug

from genericsuite.config.config import Config

from genericsuite.fastapilib.util.generic_endpoint_builder import (
    generate_blueprints_from_json
)
from genericsuite.fastapilib.endpoints import (
    users,
    menu_options,
    storage_retrieval,
)
from genericsuite.config.config_from_db import set_init_custom_data

# framework_class = importlib.import_module("fastapi")
# handler_wrapper_class = importlib.import_module("mangum")

DEBUG = False


def create_app(app_name: str, settings: Config = None) -> Any:
    """
    Create the FastAPI App
    """

    if settings is None:
        settings = Config()

    # fastapi_app = framework_class.FastAPI(title=app_name)
    fastapi_app = FastAPI(title=app_name)

    fastapi_app.debug = settings.DEBUG

    # Custom data, to be used to store a dict for
    # GenericDbHelper specific functions
    fastapi_app.custom_data = set_init_custom_data()

    # App wide log level
    # if not settings.DEBUG:
    #     fastapi_app.log.setLevel(logging.INFO)

    # CORS configuration
    set_cors_config(fastapi_app=fastapi_app, settings=settings)

    # Set Content-type: multipart/form-data as Binary
    # to properly handle image uploads
    # fastapi_app.api.binary_types.append("multipart/form-data")
    # log_debug(f'1) fastapi_app.api.binary_types: ' +
    #     '{fastapi_app.api.binary_types}')

    # Register generic endpoints
    fastapi_app.include_router(menu_options.router, prefix='/menu_options')
    fastapi_app.include_router(users.router, prefix='/users')
    fastapi_app.include_router(storage_retrieval.router, prefix='/asset')

    # Register generic endpoints (from the "endpoints.json" file)
    generate_blueprints_from_json(fastapi_app, 'endpoints')

    # Anounce app boot
    log_info(f"{settings.APP_NAME.capitalize()} v" +
             settings.APP_VERSION +
             " | Accepting connections from: " +
             settings.CORS_ORIGIN)
    if DEBUG:
        log_info(settings.debug_vars())

    return fastapi_app


def set_cors_config(fastapi_app, settings):
    """
    Sets the CORS configuration for the API.
    """
    origins = [settings.CORS_ORIGIN]
    fastapi_app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
        expose_headers=["*"],
    )


def create_handler(app_object):
    """
    Returns the FastAPI App as a valid AWS Lambda Function handler
    """
    # return handler_wrapper_class.Mangum(app_object, lifespan="off")
    return Mangum(app_object, lifespan="off")



# Desc: Helper functions for config_dbdef
from typing import Optional, Union
import json
import os

from genericsuite.config.config import Config
from genericsuite.util.utilities import log_debug

DEBUG = False
settings = Config()


def get_json_def(
    json_file_name: str,
    dir_name: str = '',
    default_value: Optional[Union[dict, list, None]] = None
# ) -> dict:
) -> Union[dict, list]:
    """
    Get JSON definition from file.

    Args:
        json_file_name (str): Name of the JSON file.
        dir_name (str, optional): Directory name. Defaults to ''.
        default_value (dict, optional): Default value. Defaults to None.

    Returns:
        dict: JSON definition.
    """
    if not default_value:
        default_value = {}

    app_path = os.getcwd()
    filename = f'{app_path}/{dir_name}/{json_file_name}.json'
    if not os.path.isfile(filename):
        if DEBUG:
            log_debug(f'File {filename} not found.')
        return default_value
    with open(filename, encoding="utf-8") as json_file:
        fe_db_def = json.load(json_file)
    return fe_db_def


def get_json_def_both(json_file_name: str) -> dict:
    """Get JSON definition from both frontend and backend.

    Args:
        json_file_name (str): Name of the JSON file.

    Returns:
        dict: JSON definition.
    """
    cnf_db_base_path = settings.GIT_SUBMODULE_LOCAL_PATH
    cnf_db = get_json_def(json_file_name, f'{cnf_db_base_path}/frontend', None)
    second_json = get_json_def(json_file_name, f'{cnf_db_base_path}/backend', None)
    if second_json:
        if not cnf_db:
            cnf_db = second_json.copy()
        elif isinstance(cnf_db, dict):
            cnf_db.update(second_json)
        else:
            cnf_db += second_json
    if not cnf_db:
        cnf_db = {}
    # log_debug(f'>>> CNF_DB\n| json_file_name: {json_file_name}\n| cnf_db: {cnf_db}')
    return cnf_db



"""
FastAPI parse multipart module to handle downloaded files.
"""
from typing import Optional

from fastapi import UploadFile

from genericsuite.util.utilities import get_file_extension
from genericsuite.util.parse_multipart import download_file


async def download_file_fa(
    file: UploadFile,
    extension: Optional[str] = None
) -> str:
    """
    Downloads a file from a FastAPI UploadFile object.
    Args:
        file (UploadFile): the file to download.
    Returns:
        str: the path to the downloaded file.
    """
    if extension is None:
        extension = get_file_extension(file.filename)
    contents = await file.read()
    uploaded_file_path = download_file(file_data=contents, extension=extension)
    contents = None
    return uploaded_file_path



"""
This module contains utility functions for working with files.
"""
from typing import Optional
import os
from uuid import uuid4

from genericsuite.config.config import Config


def temp_filename(extension: Optional[str] = None) -> str:
    """
    Returns a temporary filename with the specified extension.

    Args:
    extension (str): The extension of the temporary file.
    If None, the extension is randomly generated.

    Returns
    str: The temporary path/filename.
    """
    settings = Config()
    if extension is None:
        filename = f"{uuid4().hex}"
    else:
        filename = f"{uuid4().hex}.{extension}"
    file_path = os.path.join(settings.TEMP_DIR, filename)
    return file_path



"""
JWT Library
"""
import base64
from typing import Callable
import datetime

import jwt
from pydantic import BaseModel

from genericsuite.util.framework_abs_layer import (
    Request, get_current_framework)

from genericsuite.util.app_logger import log_debug, log_error
from genericsuite.util.utilities import (
    standard_error_return,
    get_default_resultset,
    get_id_as_string,
)
from genericsuite.config.config import Config

settings = Config()

# ----------------------- JWT -----------------------

EXPIRATION_MINUTES = 30
DEBUG = False


class AuthTokenPayload(BaseModel):
    """
    Represents the user's payload structure of the JWT token.
    """
    public_id: str


class AuthorizedRequest(Request):
    """
    Represents the AuthorizedRequest payload structure of the JWT token,
    containing the authenticated user's data.
    """
    user: AuthTokenPayload


def request_authentication() -> Callable[[Request], AuthorizedRequest]:
    """
    Returns a function that performs request authentication with the specified
    audience list.

    Args:
        None

    Returns:
        Callable[[Request], AuthorizedRequest]: A function that performs
        request authentication.
    """
    def create_auth_request(request: Request) -> AuthorizedRequest:
        return get_general_authorized_request(request)
    return create_auth_request


def get_general_authorized_request(request: Request) -> AuthorizedRequest:
    """
    Get the authorized request from the request object.

    Args:
        request (Request): The request object.

    Returns:
        AuthorizedRequest: The authorized request.
    """
    if settings.HEADER_TOKEN_ENTRY_NAME not in request.headers:
        return standard_error_return('A valid token is missing')
    try:
        token_raw = request.headers[settings.HEADER_TOKEN_ENTRY_NAME]
        jwt_token = token_raw.replace('Bearer ', '')
        if DEBUG:
            log_debug(
                '||| REQUEST_AUTHENTICATION' +
                '\n | HEADER_TOKEN_ENTRY_NAME: ' +
                f'{settings.HEADER_TOKEN_ENTRY_NAME}' +
                f'\n | token_raw: {token_raw}' +
                f'\n | jwt_token: {jwt_token}' +
                # f'\n | settings.APP_SECRET_KEY: {settings.APP_SECRET_KEY}' +
                '\n')
        jws_token_data = jwt.decode(
            jwt_token,
            settings.APP_SECRET_KEY,
            algorithms="HS256",
        )
        if DEBUG:
            log_debug(
                '||| REQUEST_AUTHENTICATION' +
                f' | jws_token_data = {jws_token_data}')

        if get_current_framework() == 'chalice':
            authorized_request = AuthorizedRequest(
                # type: ignore[attr-defined]
                event_dict=request.to_original_event(),
                # type: ignore[attr-defined]
                lambda_context=request.lambda_context,
            )
            # Authentication token data
            authorized_request.user = jws_token_data
        else:
            authorized_request = AuthorizedRequest(
                # type: ignore[attr-defined]
                event_dict=request.to_original_event(),
                # type: ignore[attr-defined]
                lambda_context=request.lambda_context,
                # Authentication token data
                user=jws_token_data,
            )

        if DEBUG:
            log_debug(
                '||| REQUEST_AUTHENTICATION' +
                f' | authorized_request = {authorized_request}')
    except Exception as err:
        log_error(
            'REQUEST_AUTHENTICATION' +
            f' | Exception = {str(err)}')
        return standard_error_return('Token is invalid')
    return authorized_request


def token_encode(user):
    """
    Encode a JWT token for the given user.

    Args:
        user: The user for whom the token is being encoded.

    Returns:
        str: The encoded JWT token.
    """
    token = jwt.encode(
        {
            'public_id': get_id_as_string(user),
            'exp':
                datetime.datetime.utcnow() +
                datetime.timedelta(minutes=EXPIRATION_MINUTES)
        },
        settings.APP_SECRET_KEY,
        algorithm="HS256"
    )
    return token


def get_basic_auth(authorization):
    """
    Decode and extract user and password from the Basic Authorization header.

    Args:
        authorization (str): The Basic Authorization header.
        e.g. authorization = app.current_request.headers.get('Authorization',
                                                             '')

    Returns:
        dict: A dictionary containing the user and password.
    """
    response = get_default_resultset()
    if not authorization.startswith('Basic '):
        response['error'] = True
        response['error_message'] = 'Authorization Required'
        response['status_code'] = 401
        # return Response(body='Authorization Required', status_code=401)
        return response

    credentials = authorization[len('Basic '):]
    decoded_credentials = base64.b64decode(
        credentials.encode('ascii')
    ).decode('ascii')
    user, password = decoded_credentials.split(':', 1)
    response['resultset'] = {
        'user': user,
        'password': password,
    }
    return response



"""
Cryptography module for encrypting and decrypting strings.
"""
from cryptography.fernet import Fernet, InvalidToken


def encrypt_string(seed: str, string: str) -> str:
    """
    Encrypt a string with a seed string.

    Args:
        seed (str): Seed string.
        string (str): String to be encrypted.

    Returns:
        str: Encrypted string.
    """
    encrypted = Fernet(seed).encrypt(string.encode())
    return encrypted.decode()


def decrypt_string(seed: str, encrypted_str: str) -> str:
    """
    Decrypt an encrypted string with a seed string.

    Args:
        seed (str): Seed string.
        encrypted_str (str): Encrypted string.

    Returns:
        str: Decrypted string.
    """
    encrypted = encrypted_str.encode()
    try:
        decrypted = Fernet(seed).decrypt(encrypted)
    except InvalidToken:
        return None
    return decrypted.decode()



"""
Navigation operation helpers
"""
import json
from genericsuite.util.app_logger import log_debug
from genericsuite.util.utilities import (
    get_query_params, get_standard_base_exception_msg
)

DEBUG = False


def get_total_pages(total_rows: int, limit: int) -> int:
    """
    Calculate the total number of pages based on the total number of rows
    and the limit per page.

    Args:
        total_rows (int): The total number of rows.
        limit (int): The limit of items per page.

    Returns:
        int: The total number of pages.
    """
    if limit == 0:
        return 1
    total_pages = int(total_rows / limit)
    if total_rows % limit > 0:
        total_pages += 1
    return total_pages


def get_navigation_params(request) -> (int, int, int):
    """
    Get navigation parameters from the request query parameters.

    Args:
        request: The request object.

    Returns:
        tuple: A tuple containing limit, skip, and page.
    """
    query_params = get_query_params(request)
    limit = query_params.get('limit')
    page = query_params.get('page')
    # skip = query_params.get('skip')
    limit = int((limit, 0)[limit is None])
    page = int((page, 1)[page is None])
    # skip = int((skip, 0)[skip is None])
    skip = (page * limit) - limit
    return (limit, skip, page)


def put_total_pages_from_resultset(
    limit: int,
    resultset: dict,
) -> dict:
    """
    Calculate total pages from resultset and limit.

    Args:
        limit (int): The limit of items per page.
        resultset (dict): The resultset containing the data.

    Returns:
        dict: The resultset with totalPages added.
    """
    if DEBUG:
        log_debug('>>--> *** put_total_pages_from_resultset ***' +
                  f'\nResultset: {resultset}' +
                  f'\nLimit: {limit}')

    if resultset['error']:
        return resultset
    resultset['totalPages'] = get_total_pages(
        len(json.loads(resultset['resultset'])), limit
    )
    if DEBUG:
        log_debug(f'NEW Resultset: {resultset}')
    return resultset


def put_total_pages_in_resultset(
    db_collection,
    listing_filter: dict,
    limit: int,
    resultset: dict,
    error_code: str,
):
    """
    Calculate total pages from resultset and limit.

    Args:
        db_collection: The collection from the database.
        listing_filter: The filter to apply to the collection.
        skip: The number of documents to skip.
        limit: The limit of items per page.
        resultset: The resultset containing the data.
        error_code: The error code to use in case of an exception.

    Returns:
        dict: The resultset with totalPages added.
    """
    if resultset['error']:
        return resultset
    count_docs = db_collection.count_documents(listing_filter)
    if DEBUG:
        log_debug('PUT_TOTAL_PAGES_IN_RESULTSET' +
                  f' | count_docs: {count_docs}')
    try:
        resultset['totalPages'] = get_total_pages(
            count_docs, limit
        )
    except BaseException as err:
        resultset['error_message'] = get_standard_base_exception_msg(
            err, error_code
        )
        resultset['error'] = True
    return resultset



"""
Current user data module
"""
from typing import Any
from genericsuite.util.framework_abs_layer import get_current_framework
from genericsuite.util.generic_db_helpers import GenericDbHelper
from genericsuite.util.jwt import AuthorizedRequest
from genericsuite.util.utilities import (
    get_default_resultset,
    error_resultset,
)

DEBUG = False
NON_AUTH_REQUEST_USER_ID = "[N/A/R]"

def get_curr_user_id(request: AuthorizedRequest) -> str:
    """Get the current user ID"""
    user_id = None
    authorized_request = hasattr(request, 'user') and request.user
    if authorized_request:
        if get_current_framework() == 'chalice':
            user_id = request.user.get("public_id")
        else:
            user_id = request.user.public_id
    else:
        # Is a non-authorization request, so returns the identificator
        # 'N/A/R' meaning "Non-Authorization Request"
        user_id = NON_AUTH_REQUEST_USER_ID
    return user_id


def get_curr_user_data(
    request: AuthorizedRequest,
    blueprint: Any
) -> dict:
    """Get the current user data."""
    user_response = get_default_resultset()
    user_id = get_curr_user_id(request)
    if not user_id:
        user_response = error_resultset(
            "No user ID in authorization header", "GCUD-E010")
    elif user_id == NON_AUTH_REQUEST_USER_ID:
        # Is a non-authorization request, so returns
        # the 'resultset' as a empty dict and no error
        pass
    else:
        dbo = GenericDbHelper(json_file="users", request=request, blueprint=blueprint)
        user_response = dbo.fetch_row_raw(user_id, {'passcode': 0})
    return user_response



"""
Schema utilities
"""
from typing import Union
from logging import Logger
from marshmallow import Schema, ValidationError


def schema_verification(
    json_body: dict,
    schema_validator: Schema,
    app_logger: Logger,
) -> Union[dict, None]:
    """
    Validate the input data against the provided schema.

    Args:
        json_body (dict): The input data to be validated.
        schema (Schema): The schema to validate input data.
        app_logger (Logger): The logger to log validation errors.

    Returns:
        Union[dict, None]: The validated data or None if the input data
            does not conform to the schema.
    """
    try:
        return schema_validator.load(json_body)
    except ValidationError as error:
        app_logger.error(f'Query error: {error.messages}')
    return None



"""
FastAPI abstraction layer
"""
from typing import Optional, Union, Dict, Any

import os
import importlib
import json

# from fastapi import HTTPException
# from fastapi import Request as FastAPIRequest
# from fastapi import Blueprint as FastAPIBlueprint
from fastapi import Response as FastAPIResponse
from pydantic import BaseModel

from genericsuite.fastapilib.util.blueprint_one import (
    BlueprintOne as FaBlueprintOne
)

DEBUG = False

FRAMEWORK_LOADED = False
FRAMEWORK = os.environ.get('CURRENT_FRAMEWORK', '').lower()
if FRAMEWORK == 'fastapi':
    try:
        framework_module = importlib.import_module(FRAMEWORK)
        FRAMEWORK_LOADED = True
        if DEBUG:
            print(f'FastAPI abstraction | framework_module: {framework_module}')


        class FrameworkClass(framework_module.FastAPI):
            """
            Framkework class cloned from the selected framework super class.
            """


        class Request(BaseModel):
            """
            Request class cloned from the selected Request framework super class.
            This class is the one to be imported by the project modules
            """
            method: Optional[str] = "GET"
            query_params: Optional[dict] = {}
            json_body: Optional[dict] = {}
            headers: Optional[dict] = {}
            event_dict: Optional[Dict[str, Any]] = {}
            lambda_context: Optional[Any] = None

            def to_dict(self):
                """
                Returns the request data as a dictionary.
                """
                return {
                    "method": self.method,
                    "query_params": self.query_params,
                    "json_body": self.json_body,
                    "headers": self.headers,
                }

            def to_original_event(self) -> Union[Dict[str, Any], None]:
                """
                Returns the original event dictionary.
                """
                return self.event_dict


        class Response(FastAPIResponse):
            """
            Response class cloned from the selected Response framework super class.
            This class is the one to be imported by the project modules
            """
            body: Union[str, dict]
            status_code: Optional[int] = 200
            headers: Optional[dict] = {}

            def __init__(
                self,
                body: Union[str, dict],
                status_code: Optional[int] = 200,
                headers: Optional[dict] = None
            ):
                """
                Initializes the Response object.
                """
                if isinstance(body, dict):
                    body = json.dumps(body)

                headers = headers if headers else {}
                if 'Content-Type' not in headers:
                    headers['Content-Type'] = 'application/json'
                if 'Access-Control-Allow-Origin' not in headers:
                    headers["Access-Control-Allow-Origin"] = os.environ.get('APP_CORS_ORIGIN', '*')
                if 'Access-Control-Allow-Methods' not in headers:
                    headers["Access-Control-Allow-Methods"] = "GET, POST, PUT, DELETE, OPTIONS"
                if 'Access-Control-Allow-Headers' not in headers:
                    headers["Access-Control-Allow-Headers"] = "Content-Type, Authorization"

                if DEBUG:
                    print('FastAPI abstraction' +
                          f'\n| body: {body}' +
                          f'\n| status_code: {status_code}' +
                          f'\n| headers: {headers}')
                super().__init__(
                    content=body,
                    status_code=status_code,
                    headers=headers
                )


        class Blueprint(framework_module.APIRouter):
            """
            Blueprint class cloned from the selected Blueprint framework super class.
            This class is the one to be imported by the project modules
            """

        class BlueprintOne(FaBlueprintOne):
            """
            Class to register a new route with optional schema validation and authorization.
            """

    except ImportError as err:
        raise ImportError(f"Unable to import '{FRAMEWORK}': {err}") from None



"""
Logging utilities
"""

from typing import Any
import os
import sys
import logging
import logging.config
import datetime

from genericsuite.config.config import Config, is_local_service

settings = Config()


def log_config() -> logging:
    """ Logging configuration """
    logger = logging.getLogger(settings.APP_NAME)
    logger.propagate = False
    if settings.DEBUG:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(name)s-%(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


app_logs = log_config()


def db_stamp() -> str:
    db_engine = os.environ['APP_DB_ENGINE']
    if db_engine == 'DYNAMO_DB':
        response = f"{db_engine}|" + \
                   f"{os.environ.get('DYNAMDB_PREFIX', 'No-Prefix')}"
    else:
        response = f"{db_engine}|{os.environ['APP_DB_NAME']}"
    if is_local_service():
        response += "|LOCAL"
    else:
        response += "|CLOUD"
    return response


def formatted_message(message: Any) -> str:
    """ Returns a formatted message with database name and date/time """
    return f"[{db_stamp()}]" + \
           f" {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}" + \
           f" | {message}"


def log_debug(message: Any) -> str:
    """Register a Debug log"""
    fmt_msg = formatted_message(message)
    app_logs.debug("%s", fmt_msg)
    return fmt_msg


def log_info(message: Any) -> str:
    """Register a Warning log"""
    fmt_msg = formatted_message(message)
    app_logs.info("%s", fmt_msg)
    return fmt_msg


def log_warning(message: Any) -> str:
    """Register a Warning log"""
    fmt_msg = formatted_message(message)
    app_logs.warning("%s", fmt_msg)
    return fmt_msg


def log_error(message: Any) -> str:
    """Register a Warning log"""
    fmt_msg = formatted_message(message)
    app_logs.error("%s", fmt_msg)
    return fmt_msg



"""
Generic endpoint operations module
"""
from bson.json_util import dumps

# from chalice.app import Response
from genericsuite.util.framework_abs_layer import Response

from genericsuite.util.app_context import AppContext
from genericsuite.util.nav_helpers import get_navigation_params
from genericsuite.util.app_logger import log_debug
from genericsuite.util.utilities import (
    return_resultset_jsonified_or_exception,
    get_query_params,
    get_request_body,
    method_not_allowed,
    get_default_resultset,
    error_resultset,
)
from genericsuite.util.generic_db_helpers import GenericDbHelper

DEBUG = False


class GenericEndpointHelper:
    """
    Helper class for generic endpoint CRUD operations.
    """
    def __init__(
        self,
        app_context: AppContext,
        url_prefix: str,
        json_file: str,
        db_type: str = None,
    ):
        self.json_file = json_file
        self.app_context = app_context
        self.request = self.app_context.get_request()
        self.query_params = get_query_params(self.request)
        self.request_body = get_request_body(self.request)
        self.data = {}
        self.data["url_prefix"] = url_prefix
        self.dbo = GenericDbHelper(json_file=json_file,
                                   request=self.request,
                                   blueprint=app_context.get_blueprint(),
                                   db_type=db_type)
        if isinstance(self.dbo.cnf_db, list):
            self.data["title"] = f'{json_file}s'
            self.data["name"] = json_file
        else:
            self.data["title"] = self.dbo.cnf_db.get('title', f'{json_file}s')
            self.data["name"] = self.dbo.cnf_db.get('name', json_file)

    def get_comb_param(self):
        """
        Get the combinator parameter.
        """
        comb_param = self.query_params.get('comb', None)
        if comb_param:
            comb_param = "$or" if comb_param in ['o', 'or'] else "$and"
        return comb_param

    def get_like_param(self):
        """
        Get the like parameter.
        """
        like_param = self.query_params.get('like', '0')
        return like_param

    def get_like_query_params(self):
        """
        Get the like query parameters.
        """
        like_query_params = {
            k: v for k, v in self.query_params.items()
            if k not in ['page', 'limit', 'like', 'comb']
        }
        return like_query_params

    def generic_crud_main(self) -> Response:
        """
        Perform generic CRUD operations.
        """
        _ = DEBUG and \
            log_debug(f'ENTERING {self.data["url_prefix"]}_crud...' +
            f' | request.method: {self.request.method}' +
            f' | query_params: {self.query_params}'
            f' | request_body: {self.request_body}')

        if self.dbo.error_message:
            return error_resultset(self.dbo.error_message)

        row_id = self.query_params.get('id')
        like_param = self.get_like_param()
        comb_param = self.get_comb_param()

        additional_query_params = self.dbo.cnf_db.get('additional_query_params')

        (limit, skip, page) = get_navigation_params(self.request)

        _ = DEBUG and log_debug(f' | row_id: {row_id}' +
            f' | additional_query_params: {additional_query_params}' +
            f' | limit: {limit} | skip: {skip} | page: {page}' + 
            f' | like_param: {like_param} | comb_param: {comb_param}' + 
            f'\n | request_body: {self.request_body}')

        if self.request.method.upper() == 'POST':
            # Create
            _ = DEBUG and \
                log_debug(f'GCM-1.1) CREATE {self.data["name"]}...')
            result = self.dbo.create_row(self.request_body)
            _ = DEBUG and \
                log_debug(f'GCM-1.2) CREATE {self.data["name"]}' +
                f'\n | request_body: {self.request_body}' +
                f'\n | result: {result}')
        elif self.request.method.upper() == 'PUT':
            # Update
            _ = DEBUG and log_debug(f'GCM-2.1) UPDATE {self.data["name"]}...')
            options = {"update_item": self.query_params.get("update_item", "0")}
            result = self.dbo.update_row(record=self.request_body,
                                         options=options)
            _ = DEBUG and log_debug(f'GCM-2.2) UPDATE {self.data["name"]}' +
                f'\n | request_body: {self.request_body}' +
                f'\n | result: {result}')
        elif self.request.method.upper() == 'DELETE' and row_id is not None:
            # Delete
            _ = DEBUG and log_debug(f'GCM-3.1) DELETE {self.data["name"]}' +
                f' | row_id: {row_id}...')
            result = self.dbo.delete_row(row_id)
            _ = DEBUG and log_debug(f'GCM-3.2) DELETE {self.data["name"]}' +
                f' | row_id: {row_id}\n | result: {result}')
        elif row_id is not None:
            # Get one row by _id
            _ = DEBUG and log_debug(f'GCM-4.1) GET ROW {self.data["name"]}' +
                f' BY ID: {row_id}...')
            result = self.dbo.fetch_row(row_id)
            _ = DEBUG and log_debug(f'GCM-4.2) GET ROW {self.data["name"]}' +
                f' BY ID: {row_id}\n | result: {result}')
        elif like_param == "1":
            # Like Search
            _ = DEBUG and log_debug(f'GCM-5.1) LIKE SEARCH {self.data["name"]}' +
                f'\n | {like_query_params}...')
            like_query_params = self.get_like_query_params()
            result = self.dbo.fetch_list(
                skip=skip,
                limit=limit,
                like_query_params=like_query_params,
                combinator=comb_param
            )
            if not result['resultset']:
                result = error_resultset(
                    f'Error: {self.data["name"]} {like_query_params} ' + \
                    "doesn't exist", "GEM3")
            _ = DEBUG and log_debug(f'GCM-5.2) LIKE SEARCH {self.data["name"]}' +
                f'\n | {like_query_params}\n | result: {result}')
        elif additional_query_params is not None and \
             all(param in self.query_params
             for param in additional_query_params):
            # Get one row by additional_query_params
            _ = DEBUG and log_debug('GCM-6.1) SEARCH BY ATTRIBUTE' +
                f' {self.data["name"]} BY {param_name}: {param_value}' +
                '...')
            result = get_default_resultset()
            param_name = None
            param_value = None
            for key in additional_query_params:
                if not self.query_params.get(key):
                    continue
                param_name = key
                param_value = self.query_params[key]
                result = self.dbo.fetch_row_by_entryname_raw(
                    entry_name=param_name,
                    entry_value=param_value,
                )
                if not result['resultset']:
                    result = error_resultset(
                        f'ERROR: {self.data["name"]} {param_value} ' + \
                        "doesn't exist", "GEM1")
                else:
                    result['resultset'] = dumps(result['resultset'])
                break
            _ = DEBUG and log_debug('GCM-6.2) SEARCH BY ATTRIBUTE' +
                f' {self.data["name"]} BY {param_name}: {param_value}' +
                f'\n | result: {result}')
        else:
            # Fetch row list
            _ = DEBUG and log_debug(f'GCM-7.1) {self.data["name"]} list |' +
                f' skip: {skip}, limit: {limit}, page: {page}' +
                f'...')
            result = self.dbo.fetch_list(skip, limit, self.query_params)
            _ = DEBUG and log_debug(f'GCM-7.2) {self.data["name"]} list |' +
                f' skip: {skip}, limit: {limit}, page: {page}' +
                f'\n | result {result}')
        return return_resultset_jsonified_or_exception(result)

    def generic_raw_json(self) -> dict:
        """
        Get the raw JSON data.
        """
        result = get_default_resultset()
        _ = DEBUG and log_debug(f'GRJ-1) {self.json_file}')
        if self.dbo.error_message:
            result['error'] = True
            result['error_message'] = self.dbo.error_message
            return result
        result['resultset'] = self.dbo.cnf_db
        # _ = DEBUG and log_debug(f'GRJ-2) {self.json_file} | result: {result}')
        return result

    def generic_json_crud(self) -> Response:
        """
        Perform generic JSON CRUD operations.
        """
        result = self.generic_raw_json()
        return return_resultset_jsonified_or_exception(result)

    def generic_array_crud(self) -> Response:
        """
        Perform generic array CRUD operations.
        """
        array_parent_id_value = self.query_params.get(self.dbo.parent_key_field)
        (limit, skip, page) = get_navigation_params(self.request)
        array_child_id_value = self.query_params.get(self.dbo.array_field_key)
        filters = {}
        if array_child_id_value is not None:
            filters[self.dbo.array_field_key] = array_child_id_value
        like_param = self.get_like_param()
        comb_param = self.get_comb_param()
        result = None
        # Create
        if self.request.method.upper() == 'POST':
            _ = DEBUG and log_debug(f'>>> {self.data["name"]}_crud | ' +
                f'PUT request_body = {self.request_body}')
            result = self.dbo.add_array_item_to_row(self.request_body)
        # Delete
        elif self.request.method.upper() == 'DELETE':
            _ = DEBUG and log_debug(f'>>> {self.data["name"]}_crud | ' +
                f'DELETE request_body = {self.request_body}')
            result = self.dbo.remove_array_item_from_row(self.request_body)
        # Modify
        elif self.request.method.upper() == 'PUT':
            # When one element needs to be modified, first remove it,
            # then add it again
            _ = DEBUG and log_debug(f'>>> {self.data["name"]}_crud | ' +
                f'POST request_body = {self.request_body}')
            remove_operation_result = \
                self.dbo.remove_array_item_from_row(self.request_body)
            if remove_operation_result['error']:
                result = remove_operation_result
            else:
                result = self.dbo.add_array_item_to_row(self.request_body)
        # List
        elif self.request.method.upper() == 'GET':
            # Get the list (paginated) or one especific element
            like_query_params = (
                self.get_like_query_params() if like_param == "1" else {})
            _ = DEBUG and log_debug(f'>>> {self.data["name"]}_crud | GET' +
                f' array_parent_id_value: {array_parent_id_value}' +
                f'\n | filters: {filters}, skip: {skip}, limit: {limit}' +
                f', page: {page}, like_query_params: {like_query_params}' +
                f' comb_param: {comb_param}')
            result = self.dbo.fetch_array_rows(
                row_id=array_parent_id_value,
                filters=filters,
                skip=skip,
                limit=limit,
                like_query_params=like_query_params,
                combinator=comb_param
            )
        if not result:
            return method_not_allowed()
        return return_resultset_jsonified_or_exception(result)



"""
Billing utilities module
"""
from genericsuite.config.config import Config
from genericsuite.util.app_context import AppContext

DEFAULT_PLAN = "free"
DEBUG = False


class BillingUtilities:
    """
    Billing utilities class
    """
    def __init__(self, app_context: AppContext) -> None:
        self.app_context = app_context
        self.settings = Config(app_context)
        self.user_plan = self.get_user_plan()

    def get_user_plan(self) -> str:
        """
        Get user's current billing plan
        """
        user_plan = self.app_context.get_user_data().get("plan", DEFAULT_PLAN)
        if str(user_plan).strip().lower() not in ["free", "basic", "premium"]:
            user_plan = DEFAULT_PLAN
        return user_plan

    def check_plan(self, plan_type: str) -> bool:
        """
        Returns True if the current user plan is equal to supplied plan
        """
        return self.user_plan == plan_type

    def is_free_plan(self) -> bool:
        """
        Returns True if the current user has a free plan
        """
        return self.check_plan("free")

    def is_basic_plan(self) -> bool:
        """
        Returns True if the current user has a basic plan
        """
        return self.check_plan("basic")

    def is_premium_plan(self) -> bool:
        """
        Returns True if the current user has a premium plan
        """
        return self.check_plan("premium")



"""
This module contains functions for interacting with AWS Secrets Manager.
IMPORTANT:
* It cannot use configs.py because it is used by config_from_db.py to
* It cannot use app_context, to avoid cycling imports.
"""
from typing import Callable
import os
import json

import boto3
from botocore.exceptions import ClientError


DEBUG = False
TEMP_DIR = os.environ.get('TEMP_DIR', '/tmp')


def get_secrets(secret_name: str, region_name: str,
                get_default_resultset: Callable, logger: Callable) -> dict:
    """
    Get a secret from AWS Secrets Manager.
    """
    _ = DEBUG and logger.debug(f'AWS get_secrets | secret_name: {secret_name}'
                               f' | region_name: {region_name}')
    result = get_default_resultset()
    session = boto3.session.Session()
    client = session.client(
        service_name='secretsmanager',
        region_name=region_name
    )
    try:
        get_secret_value_response = client.get_secret_value(
            SecretId=secret_name
        )
    except ClientError as e:
        # For a list of exceptions thrown, see
        # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
        result['error'] = True
        result['error_message'] = str(e) + ' [A-ACF-E030]'
        result['secret_name'] = secret_name
        result['region_name'] = region_name
        return result
    secret = get_secret_value_response['SecretString']
    # _ = DEBUG and logger.debug(f'get_secret | secret: {secret}')
    try:
        result['resultset'] = json.loads(secret)
    except ValueError as e:
        result['error'] = True
        result['error_message'] = str(e) + ' [A-ACF-E040]'
    except Exception as e:
        result['error'] = True
        result['error_message'] = str(e) + ' [A-ACF-E050]'
    return result


def get_secrets_cache_filename(secret_type: str = "") -> str:
    """
    Get the filename for the secrets cache.
    """
    app_name = os.environ.get('APP_NAME')
    app_stage = os.environ.get('APP_STAGE')
    if not app_name or not app_stage:
        error_message = 'ERROR: Missing environment variables' + \
            ' APP_NAME, APP_STAGE [A-ACF-E010]'
        raise Exception(error_message)
    prefix = {
        'secrets': 's_ec',
        'envs': 'e_nv'
    }
    if secret_type not in ['secrets', 'envs']:
        error_message = f'ERROR: Invalid secret_type: {secret_type}' + \
            ' [A-ACF-E011]'
        raise Exception(error_message)
    return os.path.join(
        TEMP_DIR, f'{prefix[secret_type]}_{app_name.lower()}_' +
                  f'{app_stage.lower()}_aws.json')


def get_cache_secret(get_default_resultset: Callable, logger: Callable
                     ) -> dict:
    """
    Try to get the secrets from the secrets cache file.
    If it doesn't exist, create it from AWS Secrets.
    """
    app_name = os.environ.get('APP_NAME')
    app_stage = os.environ.get('APP_STAGE')
    region_name = os.environ.get('AWS_REGION')
    result = get_default_resultset()
    if not app_name or not app_stage or not region_name:
        result['error'] = True
        result['error_message'] = 'ERROR: Missing environment variables' + \
            ' (APP_NAME, APP_STAGE, AWS_REGION) [A-ACF-E020]'
        return result
    secret_sets = []
    if os.environ.get("GET_SECRETS_CRITICAL", "1") == "1":
        _ = DEBUG and logger.debug("GET_SECRETS_CRITICAL set to 1..." +
                                   " getting secrets from the cloud...")
        secret_sets.append({
            "encrypted": True,
            "secret_name": f'{app_name.lower()}-{app_stage.lower()}-secrets',
            "secrets_cache_filename": get_secrets_cache_filename('secrets')
        })
    else:
        _ = DEBUG and logger.debug("GET_SECRETS_CRITICAL set to 0..." +
                                   " getting secrets from environment...")
    if os.environ.get("GET_SECRETS_ENVVARS", "1") == "1":
        _ = DEBUG and logger.debug("GET_SECRETS_ENVVARS set to 1..." +
                                   " getting envvars from the cloud...")
        secret_sets.append({
            "encrypted": False,
            "secret_name": f'{app_name.lower()}-{app_stage.lower()}-envs',
            "secrets_cache_filename": get_secrets_cache_filename('envs')
        })
    else:
        _ = DEBUG and logger.debug("GET_SECRETS_ENVVARS set to 0..." +
                                   " getting envvars from environment...")
    result['resultset'] = {}
    for secret_set in secret_sets:
        secret_name = secret_set["secret_name"]
        secrets_cache_filename = secret_set["secrets_cache_filename"]
        if os.path.exists(secrets_cache_filename):
            _ = DEBUG and logger.debug(
                f'AWS get_cache_secret | secret_name: {secret_name}'
                f' | FROM secrets_cache_filename: {secrets_cache_filename}')
            with open(secrets_cache_filename, 'r', encoding="utf-8") as f:
                result['resultset'].update(json.load(f))
        else:
            _ = DEBUG and logger.debug(
                f'AWS get_cache_secret | secret_name: {secret_name}' +
                f' | region_name: {region_name}' +
                f' | CREATING cache filename: {secrets_cache_filename}')
            result_inner = get_secrets(
                secret_name, region_name,
                get_default_resultset, logger)
            if result_inner['error']:
                result['error'] = True
                result['error_message'] = result_inner['error_message']
            else:
                with open(secrets_cache_filename, 'w', encoding="utf-8") as f:
                    json.dump(result_inner['resultset'], f)
                result['resultset'].update(result_inner['resultset'])
    return result



"""
General Utilities
"""
from typing import Optional, Match, AnyStr, Any, Union
import os
import re
import sys
import traceback
import base64

from mimetypes import MimeTypes

# from flask import jsonify, make_response
# from flask_cors import cross_origin

# from chalice.app import Response, Request
from genericsuite.util.framework_abs_layer import Response, Request

from genericsuite.util.app_logger import log_error, log_debug

# Email validation regular expression
EMAIL_REGEX = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b'

# send_file() mode
SEND_FILE_AS_BINARY = False

DEBUG = False


def check_email(email) -> Optional[Match[AnyStr]]:
    """ Check that an email address is valid"""
    return re.fullmatch(EMAIL_REGEX, email)


def email_verification(data, email_fields):
    """
    Verifies the email fields in the data.

    Args:
        data (dict): The data containing the email fields to be verified.
        email_fields (list): The list of email fields to be verified.

    Returns:
        dict: The resultset containing the error message if any
        email field is invalid.
    """
    resultset = get_default_resultset()
    for email_field in email_fields:
        if email_field not in data:
            resultset['error_message'] = \
                'error: Email wasn\'t specified [EMV1].'
            break
        if data[email_field] == 'foo@baz.com' \
                and not (
                    'pytest_run' in data
                    and data['pytest_run'] == 1
                ):
            resultset['error_message'] = \
                f"error: {data[email_field]} is invalid [EMV2]."
            break
        if not check_email(data[email_field]):
            resultset['error_message'] = \
                f"error: Malformed email {data[email_field]} [EMV3]."
            break
    return resultset


def get_file_size(file_size, measure="mb"):
    """Returns the file size in the given measure"""
    result = f'{file_size} bytes'
    measure = measure.lower()
    if measure == "kb":
        result = f'{file_size / 1024:.2f} Kb'
    if measure == "mb":
        result = f'{file_size / (1024 * 1024):.2f} Mb'
    if measure == "gb":
        result = f'{file_size / (1024 * 1024 * 1024):.2f} Gb'
    if measure == "tb":
        result = f'{file_size / (1024 * 1024 * 1024 * 1024):.2f} Tb'
    if measure == "pb":
        result = f'{file_size / (1024 * 1024 * 1024 * 1024 * 1024):.2f} Pb'
    return result


def get_default_resultset() -> dict:
    """Returns an standard base resultset, to be used in the building
       of responses to the outside world
    """
    resultset = {
        'error': False,
        'error_message': None,
        'totalPages': None,
        'resultset': {}
    }
    return resultset


def error_resultset(
    error_message: str,
    message_code: str = ''
) -> dict:
    """
    Return an error resultset.
    """
    message_code = f" [{message_code}]" if message_code else ''
    result = get_default_resultset()
    result['error'] = True
    result['error_message'] = f"{error_message}{message_code}"
    return result


def return_resultset_jsonified_or_exception(
    result: dict,
    http_error: int = 400,
    headers: dict = None,
) -> Response:
    """Standard way to return results to to outside world.
       If there's an error, returns a header with an HTTP error code
    """
    if not headers:
        headers = {}
    if result['error'] or result['error_message']:
        if DEBUG:
            log_debug(
                'return_resultset_jsonified_or_exception |' +
                f' ERROR error_message: {result["error_message"]}' +
                f' | http_error: {http_error}'
            )
        return standard_error_return(
            error_message=result['error_message'],
            error_code=http_error,
            headers=headers,
        )
    return Response(body=result, status_code=200, headers=headers)


def get_standard_base_exception_msg(
    err: Any,
    message_code: Optional[Union[str, None]] = 'NO_E_CODE'
) -> str:
    """
    When a BaseException is fired, use this method to return
    a standard error message
    """
    message_code = f" [{message_code}]" if message_code else ''
    response = f"Unexpected Error: {err}, {type(err)}{message_code}"
    log_error(response)
    log_error(format_stacktrace())
    return response


def send_file(
    file_to_send: str,
    download_name: str = None,
) -> Response:
    """
    Send the contents of a file to the client.
    """
    if SEND_FILE_AS_BINARY:
        return send_file_as_bin(
            file_to_send=file_to_send,
            download_name=download_name,
        )
    return send_file_text_text(
        file_to_send=file_to_send,
        download_name=download_name,
    )


def send_file_as_bin(
    file_to_send: str,
    download_name: str = None,
) -> Response:
    """
    Send the contents of a file to the client as binary.
    """
    if not download_name:
        download_name = os.path.basename(file_to_send)
    # Verify file existence
    if not os.path.exists(file_to_send):
        return standard_error_return(
            error_message=f'File not found: {file_to_send}',
            error_code=404,
        )
    # Create a Chalice Response object to send the file back to the client
    with open(file_to_send, 'rb') as file_handler:
        file_data = file_handler.read()
    headers = {
        'Content-Type': get_mime_type(file_to_send),
        # 'Content-Type': 'application/octet-stream',
        'Content-Disposition': f'attachment; filename="{download_name}"',
    }
    if DEBUG:
        log_debug("SEND_FILE_AS_BIN" +
                  f"\n | file_to_send: {file_to_send}" +
                  f"\n | download_name: {download_name}" +
                  f"\n | headers: {headers}")
        # log_debug("2) SEND_FILE | bin_to_b64_to_ascii(file_data):\n" +
        #           f"{bin_to_b64_to_ascii(file_data)}")

    return Response(
        body=file_data,
        status_code=200,
        headers=headers
    )


def send_file_text_text(
    file_to_send: str,
    download_name: str = None,
) -> Response:
    """
    Send the contents of a file to the client as Base64 encoded text.
    """
    if not download_name:
        download_name = os.path.basename(file_to_send)
    # Verify file existence
    if not os.path.exists(file_to_send):
        return standard_error_return(
            error_message=f'File not found: {file_to_send}',
            error_code=404,
        )
    # Create a Chalice Response object to send the file back to the client
    with open(file_to_send, 'rb') as file_handler:
        file_data = file_handler.read()
    file_encoded = bin_to_b64_to_ascii(file_data)

    headers = {
        'Content-Type': "text/text",
        'Content-Disposition': f'attachment; filename="{download_name}"',
    }
    if DEBUG:
        log_debug("SEND_FILE_TEXT_TEXT" +
                  f"\n | file_to_send: {file_to_send}" +
                  f"\n | download_name: {download_name}" +
                  f"\n | headers: {headers}")
        log_debug(f"2) SEND_FILE |file_encoded:\n{file_encoded}")
    return Response(
        body=file_encoded,
        status_code=200,
        headers=headers
    )


def bin_to_b64_to_ascii(data):
    """
    Do the conversion that Chalice does for binary files...
    """
    _ = DEBUG and log_debug(f"BIN_TO_B64_TO_ASCII | data:\n{data}")
    data = base64.b64encode(data)
    _ = DEBUG and log_debug("BIN_TO_B64_TO_ASCII | " +
                            f"base64.b64encode(data):\n{data}")
    return data.decode('ascii')


def format_stacktrace() -> str:
    """Returns the stack trace summarized

    Returns:
        str: the last 2 files and the exception message
             from the stack trace
    """
    parts = ["Traceback (most recent call last):\n"]
    parts.extend(traceback.format_stack(limit=25)[:-2])
    parts.extend(traceback.format_exception(*sys.exc_info())[1:])
    return "".join(parts)


# @cross_origin(supports_credentials=True)
def standard_error_return(
    error_message: str,
    error_code: int = 401,  # Unauthorized
    headers: dict = None
) -> Response:
    """Returns a standard error response"""
    if not headers:
        headers = {'WWW.Authentication': 'Basic realm: "login required"'}
    # return make_response(error_message, error_code, headers)
    return Response(
        body=error_message, status_code=error_code, headers=headers
    )


def method_not_allowed() -> Response:
    """Return a response with 405 Method Not Allowed"""
    return Response(body='Method Not Allowed', status_code=405)


def get_query_params(request: Request) -> dict:
    """Returns the query parameters (Chalice)"""
    query_params = request.to_dict()['query_params']
    if query_params is None:
        query_params = {}
    return query_params


def get_request_body(request: Request) -> dict:
    """Returns the request body (Chalice)"""
    try:
        request_body = request.json_body
    except BaseException:
        request_body = dict()
    return request_body


def is_an_url(element_url_or_path: str):
    """ Returns True if the string is an URL"""
    return element_url_or_path.startswith(
        ("http://", "https://", "ftp://", "file://")
    )


def sort_list_of_dicts(data: list[dict], column_name: str, direction: str):
    """
    Sorts a list of dictionaries by the specified column name and direction.

    Args:
        data (list[dict]): The list of dictionaries to be sorted.
        column_name (str): The name of the column to sort by.
        direction (str): The direction of the sort, either "asc" or "desc".

    Returns:
        list[dict]: The sorted list of dictionaries.
    """

    if DEBUG:
        log_debug("")
        log_debug(f"sort_list_of_dicts | data: {data}")
    sorted_data = sorted(
        data,
        key=lambda x: str(x.get(column_name)).upper(),
        reverse=direction == "desc"
    )
    if DEBUG:
        log_debug(f"sort_list_of_dicts | sorted_data: {sorted_data}")
        log_debug("")
    return sorted_data


# Previouslly: get_user_id_as_string
def get_id_as_string(row):
    """
    Returns the Object(_ID) of a row as a string.

    Args:
        row (dict): The row containing the ID (MongoDb).

    Returns:
        str: The ID of the row as a string.
    """
    return str(row['_id'])


def get_default_value(entryname, dict_obj, default_value):
    """
    Try to get the entryname in the dict, if not exist
    returns the default value.
    """
    if entryname in dict_obj:
        result = dict_obj[entryname]
    else:
        result = default_value
    return result


def get_mime_type(file_path: str) -> str:
    """
    Get the MIME type for a given file extension.

    Args:
        file_path (str): file name, file path or extension only.

    Returns:
        str: The corresponding mimetype (e.g., 'image/jpeg', 'image/png').
    """
    file_path_to_guess_type = file_path
    extension = get_file_extension(file_path=file_path)
    if extension == '':
        # Assumes it's a extension-only, not a file name
        extension = file_path
        file_path_to_guess_type = f"dummy.{extension}"
    mime = MimeTypes()
    mime_type, _ = mime.guess_type(file_path_to_guess_type)
    if mime_type:
        return mime_type
    mime_types = {
        'wav': 'audio/x-wav',
        'mp3': 'audio/mpeg',
        'opus': 'audio/opus',
        'jpg': 'image/jpeg',
        'jpeg': 'image/jpeg',
        'png': 'image/png',
        'gif': 'image/gif',
        'bmp': 'image/bmp',
        'tiff': 'image/tiff',
        'svg': 'image/svg+xml',
    }
    return mime_types.get(extension.lower(), 'application/octet-stream')


def get_valid_extensions(extension_type: str = None) -> list:
    """
    Returns the list of valid extension for a given extension type.
    If the extension type is None, returns all valid extension.
    If the extension type is not valid, returns no valid extensions
    as an empty list.
    """
    valid_extensions = {
        "image": ["jpg", "jpeg", "png", "gif", "bmp", "tiff", "svg"],
        "json": ["json"],
        "text": ["txt", "md"],
        "csv": ["csv"],
        "audio": ["mp3", "wav", "flac"],
        "data_science": ["csv", "tsv", "parquet", "feather", "hdf5"],
        "other": ["docx", "pptx", "xlsx", "pdf"]
    }
    if extension_type:
        return valid_extensions.get(extension_type, [])
    flatten_list = []
    for list_element in valid_extensions.values():
        flatten_list += list_element
    return flatten_list


def get_url_query_args(url: str) -> dict:
    """
    Returns a dict with the key=value pairs in a URL query string
    """
    if len(url.split("?")) > 1:
        _, query_args_plain = url.split("?")
    else:
        query_args_plain = url
    query_args = query_args_plain.split("&")
    return {
        v.split("=")[0]: v.split("=")[1]
        if len(v.split("=")) > 1 else None
        for v in query_args
    }


def get_file_extension(file_path: str) -> str:
    """
    Returns the file extension from a path or filename.
    If there's no "." in file_path, assumes there's no file extension, so
    returns an empty string.
    """
    if file_path is None:
        return ''
    if len(file_path.split('.')) > 1:
        return file_path.split('.')[-1]
    return ''


def get_last_url_element(url: str) -> str:
    """
    Returns the last element from a URL (normally a file or page name)
    """
    return url.split('/')[-1]


def deduce_filename_from_url(url: str, extension_type: str = None) -> str:
    """
    Deduces a filename from a URL and checks if it has a valid extension.
    If not, tries to get a query parameter or REST API element that seems
    to be a filename.
    As a third attempt, uses regex to figure out a filename with a
    valid extension.
    """
    # First attempt: get the last element from the URL
    valid_extensions = get_valid_extensions(extension_type)
    filename = get_last_url_element(url)
    file_extension = get_file_extension(filename)
    if file_extension.lower() in valid_extensions:
        if DEBUG:
            log_debug(f"\n1st attempt: {filename}\n")
        return get_last_url_element(filename)
    # Second attempt: Check query parameters for a filename
    query_args = get_url_query_args(url)
    for key, value in query_args.items():
        if key.lower() in ['filename', 'file', 'name', 'f', 'n'] and \
           get_file_extension(value) in valid_extensions:
            if DEBUG:
                log_debug(f"\n2nd attempt [1]: {value} in key: {key}\n")
            return get_last_url_element(value)
    # for value in query_args.values():
    for key, value in query_args.items():
        if get_file_extension(value) in valid_extensions:
            if DEBUG:
                log_debug(f"\n2nd attempt [2]: {value} in key: {key}\n")
            return get_last_url_element(value)
    # Third attempt: Use regex to find a filename with a valid extension
    # e.g. on REST API elements
    pattern = r'/([^/?#]+)\.(' + '|'.join(valid_extensions) + ')(?=[?#]|$)'
    match = re.search(pattern, url)
    if match:
        value = match.group(1) + '.' + match.group(2)
        log_debug(f"\n3rd attempt: {value} | match: {match}\n")
        return get_last_url_element(value)
    if DEBUG:
        log_debug("\nNo way Jose\n")
    return ''


def interpret_any(any_input: Any) -> str:
    """
    Interpret the input and return a string representation.

    Args:
        any_input (Any): The input to interpret.

    Returns:
        str: A string representation of the interpreted input.
    """
    result = any_input
    if isinstance(any_input, dict):
        result = " ".join(list(any_input.values()))
    elif isinstance(any_input, list):
        result = " ".join(any_input)
    elif not isinstance(any_input, str):
        result = str(any_input)
    return result


def is_under_test() -> bool:
    """
    Returns True if the current environment is under test.
    """
    return 'PYTEST_CURRENT_TEST' in os.environ



"""
Password Encryption module
"""
from werkzeug.security import generate_password_hash, check_password_hash

from genericsuite.config.config import Config

settings = Config()


class Passwords:
    """ Class to handle all passwords operations """

    def encrypt_password(self, passcode: str) -> str:
        """
        Encrypts a password using the 'scrypt' method.
        :param passcode: The password to encrypt.
        :return: The encrypted password.
        """
        return generate_password_hash(
            settings.APP_SECRET_KEY + passcode,
            method='scrypt',
        )

    def verify_password(self, db_user_password: str,
                        form_auth_password: str) -> bool:
        """
        Verifies a password by comparing the encrypted form of
        the form_auth_password with the db_user_password.
        :param db_user_password: The encrypted password stored in the database.
        :param form_auth_password: The password entered by the user.
        :return: True if the passwords match, False otherwise.
        """
        return check_password_hash(
            db_user_password,
            settings.APP_SECRET_KEY + form_auth_password
        )

    def passwords_encryption(self, data: dict, password_fields: list) -> dict:
        """
        Apply encryption to fields listed as password fields.

        Args:
            datad (dict): The original record containing the unencrypted
            password fields.
            password_fields (list): The list of password fields to
            be encrypted.

        Returns:
            dict: The updated record with encrypted password fields.
        """
        for field in password_fields:
            if field in data:
                data[field] = self.encrypt_password(data[field])
        return data



"""
generic_endpoint_builder: generate endpoint from a json file.
"""
from pprint import pprint

# from chalice import Chalice
from genericsuite.util.framework_abs_layer import FrameworkClass as Chalice

# from chalice.app import Response
# from genericsuite.util.blueprint_one import BlueprintOne
from genericsuite.util.framework_abs_layer import Response, BlueprintOne

from genericsuite.util.config_dbdef_helpers import get_json_def
from genericsuite.util.generic_endpoint_helpers import GenericEndpointHelper
from genericsuite.util.jwt import (
    AuthorizedRequest,
    request_authentication
)
from genericsuite.util.app_logger import log_debug
from genericsuite.util.utilities import return_resultset_jsonified_or_exception
from genericsuite.config.config_from_db import app_context_and_set_env

from genericsuite.config.config import Config

DEBUG = False


def generate_blueprints_from_json(
    app: Chalice,
    json_file: str = "endpoints",
) -> None:
    """
    Generates blueprints from a JSON file and registers them with
    the Chalice app.

    Args:
        app (Chalice): The Chalice app to register the blueprints with.
        json_file (str, optional): The JSON file containing blueprint
        definitions. Defaults to "endpoints".
    """
    settings = Config()
    cnf_db_base_path = settings.GIT_SUBMODULE_LOCAL_PATH
    definitions = get_json_def(json_file, f'{cnf_db_base_path}/backend', [])

    for definition in definitions:
        bp_name = definition['name']
        url_prefix = f"/{definition.get('url_prefix', bp_name)}"
        blueprint = BlueprintOne(bp_name)

        if DEBUG:
            log_debug(
                'GENERATE_BLUEPRINTS_FROM_JSON |' +
                f' bp_name: {bp_name}' +
                f' url_prefix: {url_prefix}'
            )

        # Add routes to the blueprint
        for route in definition['routes']:
            route_endpoint = route['endpoint']
            route_methods = route['methods']
            route_handler_type = route['handler_type']
            other_params = {
                "name": bp_name,
                "app": app
            }
            other_params["params"] = route['params']
            if route_handler_type == "GenericEndpointHelper":
                route_handler = generic_route_handler
            else:
                route_handler = route['view_function']

            if DEBUG:
                log_debug(
                    'GENERATE_BLUEPRINTS_FROM_JSON |' +
                    f' route_endpoint: {route_endpoint}' +
                    f' route_methods: {route_methods}' +
                    f' route_handler_type: {route_handler_type}' +
                    f' route_handler: {route_handler}' +
                    f' other_params: {other_params}'
                )

            blueprint.route(
                path=route_endpoint,
                authorizor=request_authentication(),
                methods=route_methods,
                other_params=other_params,
            )(generic_route_handler)

        # Register the blueprint with the Chalice app
        app.register_blueprint(
            blueprint=blueprint,
            url_prefix=url_prefix,
            name_prefix=bp_name
        )


def generic_route_handler(
    request: AuthorizedRequest,
    *args,
    **kwargs,
) -> Response:
    """
    Handles generic route requests and delegates to the appropriate
    CRUD operation based on the request parameters.

    Args:
        request (AuthorizedRequest): The authorized request object.
        kwargs (dict): Additional keyword arguments.

    Returns:
        Response: The response from the CRUD operation.
    """
    if DEBUG:
        log_debug(
            "generic_route_handler |" +
            f" | kwargs: {kwargs}" +
            f" event: {request}:"
        )
        pprint(request.to_dict())

    other_params = kwargs["other_params"]
    if DEBUG:
        log_debug(
            "generic_route_handler |" +
            f" other_params: {other_params}" +
            f" request: {request}"
        )

    # Set environment variables from the database configurations.
    bp = BlueprintOne(other_params.get('name'))
    bp.register(app=other_params['app'], options={})
    app_context = app_context_and_set_env(
        request=request,
        blueprint=bp
    )
    if app_context.has_error():
        return return_resultset_jsonified_or_exception(
            app_context.get_error_resultset()
        )

    ep_helper = GenericEndpointHelper(
        app_context=app_context,
        json_file=other_params['params']["json_file"],
        url_prefix=other_params["name"]
    )
    if ep_helper.dbo.table_type == "child_listing" \
       and ep_helper.dbo.sub_type == "array":
        return ep_helper.generic_array_crud()
    return ep_helper.generic_crud_main()



"""
Blueprint wrapper to add authorization, other data and schema validation
to requests.
"""
from typing import Any, Callable, Optional
from functools import partial, wraps

from chalice.app import Blueprint, Request
# from genericsuite.util.framework_abs_layer import Blueprint, Request
from genericsuite.util.app_logger import log_debug
from genericsuite.util.schema_utilities import Schema, schema_verification


DEBUG = False


class BlueprintOne(Blueprint):
    """
    Class to register a new route with optional schema validation and authorization.
    """

    def get_current_app(self) -> Any:
        """
        Get the current App object. It must be called inside a router function.
        """
        return self.current_app

    def get_current_request(self) -> Any:
        """
        Get the current App object. It must be called inside a router function.
        """
        return self.current_app.current_request

    def route(
        self,
        path: str,
        authorizor: Optional[Callable[[Request], Request]] = None,
        schema: Optional[Schema] = None,
        other_params: Optional[dict] = None,
        **kwargs: Any,
    ) -> Callable:
        """
        Register a new route with optional schema validation and authorization.

        Args:
            path (str): The URL path for the route.
            schema (Optional[Schema]): The schema to validate the request against.
            authorizor (Optional[Callable[[Request], Request]]): The authorization
            function.
            other_params (Optional[dict]): Additional parameters to pass to the
            route.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            Callable: The registered route function.
        """
        def route_wrapper(register: Callable, route: Callable) -> Callable:
            @wraps(route)
            def route_processor(*args: Any, **kwargs: Any) -> Callable:
                current_app = self.get_current_app()
                current_request = self.get_current_request()
                if DEBUG:
                    log_debug(
                        'Request was made to: ' +
                        f'{current_request.context.get("resourcePath", path)}, ' +
                        f'HTTP method: {current_request.method}'
                    )
                if schema:
                    schema_verification(current_request.json_body, schema, current_app.log)
                if authorizor is not None:
                    current_request = authorizor(current_request)
                    if DEBUG:
                        log_debug(
                            'RESPONSE AUTHORIZOR BlueprintOne.route_wrapper' +
                            ' | current_request:'
                        )
                        log_debug(current_request.to_dict())
                    auth_response = current_request.to_dict()
                    if 'statusCode' in auth_response \
                            and auth_response['statusCode'] != 200:
                        if DEBUG:
                            log_debug(
                                'RESPONSE AUTHORIZOR FAILED | current_request.' +
                                'statusCode'
                            )
                            log_debug(current_request.to_dict()["statusCode"])
                        return current_request

                if DEBUG:
                    log_debug('RESPONSE AUTHORIZOR OK')
                kwargs['other_params'] = other_params
                return route(current_request, *args, **kwargs)

            return register(route_processor)

        handler_final = self._create_registration_function(
            handler_type='route',
            name=kwargs.pop('name', None),
            registration_kwargs={
                'path': path,
                'kwargs': kwargs
            },
        )

        # log_debug(f'ENTERING BlueprintOne.route... {path}')
        return partial(route_wrapper, handler_final)



"""
Generic Database Helper super class
"""
from uuid import uuid4

from bson.json_util import ObjectId

from genericsuite.util.framework_abs_layer import get_current_framework
from genericsuite.util.app_logger import log_debug, log_error
from genericsuite.util.config_dbdef_helpers import get_json_def_both
from genericsuite.util.utilities import (
    get_standard_base_exception_msg,
    get_default_resultset,
    get_query_params,
    get_request_body,
)
from genericsuite.util.db_abstractor import (
    # set_db_request,
    db,
)
from genericsuite.util.passwords import Passwords


DEBUG = False


class GenericDbHelperSuper:
    """
    Generic Database Helper super class
    """
    # def __init__(self, json_file, request=None, db_type=None) -> None:
    def __init__(self, json_file, request=None, blueprint=None, db_type=None
                 ) -> None:

        # set_db_request(request)

        self.error_message = None
        self.json_file = json_file
        self.cnf_db = get_json_def_both(json_file)

        # log_debug(f">>> self.cnf_db: {self.cnf_db}")

        self.blueprint = blueprint
        self.request = request
        self.query_params = get_query_params(self.request)
        self.request_body = get_request_body(self.request)

        self.table_name = json_file
        self.name = json_file.capitalize()
        self.title = f"{json_file.capitalize()}s"
        self.table_obj = None

        self.table_type = None
        self.sub_type = None
        self.array_field = None
        self.array_field_key = None
        self.parent_key_field = None
        self.parent_key_names = []

        self.allow_duplicates = False
        self.mandatory_filters = []

        if db_type:
            self.db_type = db_type
        else:
            self.db_type = 'db'
            self.table_name = self.cnf_db['table_name']
            self.name = self.cnf_db.get('name', self.name)
            self.title = self.cnf_db.get('title', self.title)
            try:
                _ = DEBUG and \
                    log_debug(f"||| GenericDbHelper | db: {db}")
                self.table_obj = db[self.table_name]
                _ = DEBUG and \
                    log_debug("||| GenericDbHelper | self.table_obj:" +
                              f" {self.table_obj}")
            except BaseException as error:
                self.error_message = "ERROR connecting to Database:" + \
                                     f" {str(error)}"
                log_error(self.error_message)
            if not self.error_message:
                self.allow_duplicates = self.cnf_db.get('allow_duplicates',
                                                        False)
                self.mandatory_filters = self.replace_special_vars(
                    self.cnf_db.get('mandatoryFilters', {})
                )
                # table_type: could be ["main_table"] | "child_listing"
                self.table_type = self.cnf_db.get('type', "main_table")
                # table_type: could be [""] | "array"
                self.sub_type = self.cnf_db.get('subType', "")
                # Examples of array sub_type's child_listing definitions:
                # array_field = 'food_times'
                # array_field_key = 'food_moment_id'
                # parent_key_field = 'user_id'
                if self.table_type == "child_listing" and \
                   self.sub_type == "array":
                    self.array_field = self.cnf_db.get('array_name')
                    self.array_field_key = self.cnf_db['primaryKeyName']
                    self.parent_key_names = self.cnf_db['parentKeyNames']
                    self.parent_key_field = \
                        self.cnf_db['parentKeyNames'][0]['parameterName']

    def listing_projection_exclusions(self) -> dict:
        """
        This method returns the projection dictionary for excluding fields
        from the listing resultset.

        Returns:
            dict: The projection dictionary for excluding fields.
            The dictionary keys are the field names and the values are 0.
            e.g. {'field_name': 0, 'field_name2': 0}
            e.g. {} (empty dictionary)
        """
        projection = {k: 0 for k in self.cnf_db.get(
            'projection_exclusion', []
        )}
        if DEBUG:
            log_debug('listing_projection_exclusions |' +
                      f' projection: {projection}')
        return projection

    def listing_disabled_columns_projection(self) -> dict:
        """
        This method returns the projection dictionary for fields
        that have no "listing": True in the listing configuration
        or the column is included in the projection_exclusion list.
        so they will be excluded from the database query result.

        Returns:
            dict: The projection dictionary for the listing resultset.
            The dictionary keys are the field names and the values are 0.
            e.g. {'field_name': 0, 'field_name2': 0}
            e.g. {} (empty dictionary)
        """
        projection = {
            k["name"]: 0 for k in self.cnf_db.get('fieldElements', [])
            if not k.get("listing", False) or
            k["name"] in self.cnf_db.get('projection_exclusion', [])
        }
        if DEBUG:
            log_debug('listing_disabled_columns_projection |' +
                      f' projection: {projection}')
        return projection

    def add_mandatory_filters(
        self,
        listing_filter: dict,
        combinator: str = None
    ) -> dict:
        """
        This method adds mandatory filters to the listing filter
        and returns the updated filter. Mandatory filters will be
        ALWAYS included as equality (==) filter with the $and
        combinator in each condition, whether it's a "like" or
        normal condition). E.g.

        self.mandatory_filters:{'user_id': 'XXXX'}

        listing_filter BEFORE: {'meal_date': {'$lte': 946702800.0,
        '$gte': 946616400.0}, 'observations': {'$regex':
        '.*sancocho.*', '$options': 'si'}, 'user_id': 'XXXX'}

        mandatory_filter_to_add: [{'user_id': 'XXXX'}]

        listing_filter AFTER: {'$or': [{'$and': [{'user_id': 'XXXX'},
        {'meal_date': {'$lte': 946702800.0, '$gte': 946616400.0}}]},
        {'$and': [{'user_id': 'XXXX'}, {'observations':
        {'$regex': '.*sanncocho.*', '$options': 'si'}}]}]}

        Args:
            listing_filter (dict): The original listing filter.
            combinator (str): The condition combinator for the filter.
            Defaults to "$and".

        Returns:
            dict: The updated listing filter with mandatory filters added.
        """
        if not combinator:
            combinator = "$and"
        if DEBUG:
            log_debug("add_mandatory_filters | self.mandatory_filters:" +
                      f"{self.mandatory_filters}")
            log_debug("add_mandatory_filters | listing_filter BEFORE:" +
                      f" {listing_filter}")
        if self.mandatory_filters.keys() != listing_filter.keys():
            # Mandatory filters will be ALWAYS included as
            # equality (==) filter with the $and combinator
            # in each "like" condition
            mandatory_filter_to_add = [
                {k: v} for k, v in self.mandatory_filters.items()
            ]
            # Filter multiple "like" conditions with the
            # specified combinator "$and" or "$or"
            if len(listing_filter) == 0 and len(mandatory_filter_to_add) == 0:
                listing_filter = {}
            elif len(listing_filter) == 0:
                listing_filter = {
                    "$and": [
                        {k: v} for k, v in self.mandatory_filters.items()
                    ]
                }
            else:
                listing_filter = {
                    combinator: [
                        {k: v} if len(mandatory_filter_to_add) == 0
                        else {'$and': list(mandatory_filter_to_add + [{k: v}])}
                        for k, v in listing_filter.items()
                        if k not in self.mandatory_filters
                    ]
                }
            if DEBUG:
                log_debug("add_mandatory_filters | mandatory_filter_to_add:" +
                          f" {mandatory_filter_to_add}")
        if DEBUG:
            log_debug("add_mandatory_filters | listing_filter AFTER:" +
                      f" {listing_filter}")
        return listing_filter

    def passwords_encryption_on_update(
        self,
        record: dict,
        updated_record: dict,
        password_fields: list,
    ) -> dict:
        """
        Encrypts password fields in the updated record based on the original
        record.

        Args:
            record (dict): The original record containing the unencrypted
            password fields.
            updated_record (dict): The updated record where the password
            fields will be encrypted.
            password_fields (list): The list of password fields to be
            encrypted.

        Returns:
            dict: The resultset containing the updated record with encrypted
            password fields.
        """
        psw_class = Passwords()
        if DEBUG:
            log_debug('BEGIN passwords_encryption_on_update | ' +
                      f'password_fields: {password_fields} | ' +
                      f'record: {record}')
        resultset = get_default_resultset()
        existing_row = None
        for field in password_fields:
            if field in record and record[field]:
                updated_record[field] = psw_class.encrypt_password(
                    record[field]
                )
            else:
                if existing_row is None:
                    existing_row = self.fetch_row_raw(
                        record['_id'], {field: 1}
                    )
                    if existing_row['error']:
                        resultset['error_message'] = \
                            existing_row['error_message']
                        resultset['error'] = True
                        return resultset
                updated_record[field] = \
                    existing_row['resultset'][field]
        resultset['updated_record'] = updated_record
        if DEBUG:
            log_debug('END passwords_encryption_on_update | ' +
                      f'resultset: {resultset}')
        return resultset

    def get_mandatory_fields(self, record: dict,
                             is_create: bool = False) -> list:
        """
        Returns the list of mandatory fields names existing in the
        given record, based on the required condition in the
        fieldElements and mandatory_fields configurations.

        Args:
            record (dict): record which mandatory fields are to be determined.
            is_create: bool, optional
                Indicates if the undergoing action is creation.
                Default is False.
                This condition is needed to handle the "_id" element:
                if "_id" is required and ...
                "_id" is in record and is creation or
                "_id" is not in record and is not creation
                then "_id" will be included as mandatory field.

        Returns:
            list
                A list of mandatory fields for the record.
        """
        mandatory_fields = ['creation_date', 'update_date']
        mandatory_fields.extend(self.cnf_db.get('mandatory_fields', []))
        for element in self.cnf_db['fieldElements']:
            element_name = element['name']
            if not element.get('required', False):
                continue
            if element.get('required', False) \
               and element.get('type', "") == "_id":
                if "_id" in record and not is_create:
                    continue
                if "_id" not in record and is_create:
                    continue
            if element_name not in mandatory_fields:
                mandatory_fields.append(element_name)
        return mandatory_fields

    def get_current_user(self) -> str:
        """
        Get the current user Id from the Request

        Returns:
            str: the current user Id from the JWT Request
        """
        if get_current_framework() == 'chalice':
            return self.request.user["public_id"]
        else:
            return self.request.user.public_id

    def replace_special_vars(self, params: dict) -> dict:
        """
        Returns the params record replacing special value tokens.
        E.g. {CurrentUserId} will be replaced with the current user Id.

        Args:
            params (dict): record (item data) to be processed.
                E.g. {'name': 'John', 'age': '30', 'gender': 'male'}
                or {'name': 'John', 'age': '30', 'gender': '{CurrentUserId}'}
        Returns:
            dict: record (item data) replacing special value tokens.
        """
        response = {
            k: self.get_current_user() if v == "{CurrentUserId}" else v
            for k, v in params.items()
        }
        # if DEBUG:
        #     log_debug("replace_special_vars | params:" +
        #               f" {params} | response: {response}")
        return response

    def get_field_element(self, fieldname: str) -> dict:
        """
        Get one column configuration from 'fieldElements'

        Args:
            fieldname (str): field name.

        Returns:
            dict: field (attribute) definition or an empty dict
                if the field is not found.
        """
        # if DEBUG:
        #     log_debug(f">>> get_field_element | fieldname: {fieldname}" +
        #               f" | fieldElements: {self.cnf_db.get('fieldElements')}"
        #               )
        field_element = [v for v in self.cnf_db.get('fieldElements', [])
                         if v.get('name') == fieldname]
        # Returns an empty dict if the field is not found
        # to handle eventual invalid fieldnames passed in the endpoint call.
        return field_element[0] if field_element else {}

    def is_datefield(self, fieldname: str) -> bool:
        """
        Determine if a column is a date.

        Args:
            fieldname (str): field name.

        Returns:
            bool: True if the column is a date, False otherwise.
        """
        return self.get_field_element(fieldname).get('type') \
            in ['date', 'datetime-local']

    def uuid_generator_check(self, data: dict, key_fieldname: str,
                             array_field: str) -> dict:
        """
        Checks if the key_fieldname has a "uuid_generator" attribute
        in the fieldElements configuration. If it does, and the
        key_fieldname is not present in the data[array_field] or
        if it is present but empty, it generates a UUID4 and assigns
        it to data[array_field][key_fieldname].

        IMPORTANT: data[array_field] must have only ONE the item,
        the one that ID (key_fieldname) will be assigned.

        Args:
            data (dict): The JSON data to be checked and updated.
            key_fieldname (str): The field name to be checked for
            UUID generation.
            array_field (str): The field name of the array to be checked
            and updated.

        Returns:
            dict: The updated JSON data with UUID4 assigned if necessary.
        """
        field_element = self.get_field_element(key_fieldname)
        if 'uuid_generator' not in field_element:
            return data
        if ((key_fieldname not in data[array_field] or
             (key_fieldname in data[array_field] and
              not data[array_field][key_fieldname]))):
            data[array_field][key_fieldname] = str(uuid4())
        return data

    def get_sort_config(self, order_param: str = None) -> (str, str):
        """
        Returns the column name and direction for sorting based on the
        query parameter "order" or default configuration
        "primaryKeyName", if not, assumes "_id".

        Returns:
            (str, str): Tuple with the column name and direction for sorting.
        """
        if order_param:
            order_str = order_param
        elif self.query_params.get('order'):
            order_str = self.query_params.get('order')
        else:
            order_str = self.cnf_db.get(
                'defaultOrder',
                self.cnf_db.get('primaryKeyName', "_id")
            )
        if "|" not in order_str:
            order_str += "|asc"
        [column_name, direction] = order_str.split("|")

        if DEBUG:
            log_debug("")
            log_debug(f">> GET_SORT_CONFIG | column_name: {column_name}," +
                      f" direction: {direction}")
            log_debug("")

        return column_name, direction

    def run_specific_func(self, action, resultset):
        """
        Runs the specific function for the given action.

        Args:
            action (str): The action to run the specific function for.
            resultset (dict): The resultset to be passed to the specific
            function.

        Returns:
            dict: The resultset returned by the specific function.
        """
        specific_func_name = self.cnf_db.get('specific_function', None)
        if specific_func_name:
            specific_func = self.blueprint.get_current_app().custom_data.get(
                specific_func_name
            )
            if not specific_func:
                raise Exception(f"Specific function {specific_func_name}" +
                                " not found.")
            action_data = {
                'action': action,
                'resultset': resultset,
                'cnf_db': self.cnf_db,
            }
            resultset = specific_func(self.blueprint, action_data)
        return resultset

    # ----- Normal row operations.

    def fetch_row_raw(self, row_id: str, projection: dict = None) -> dict:
        """
        Fetches a row from the database based on the given row_id
        and returns it without applying dumps() to the 'resultset'
        element.

        Args:
            row_id (str): The ID of the row to fetch.
            projection (dict, optional): The projection to apply to the
            fetched row.

        Returns:
            dict: The resultset containing the fetched row.
        """
        resultset = get_default_resultset()
        if not projection:
            projection = {}

        try:
            str_id = ObjectId(row_id)
        except ValueError:
            resultset['error_message'] = \
                f'Id `{row_id}` is invalid [FUR1].'
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'FUR2')
            # raise

        if resultset['error_message']:
            resultset['error'] = True
            return resultset

        try:
            resultset['resultset'] = self.table_obj.find_one(
                {'_id': str_id}, projection
            )
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'FUR3')
            resultset['error'] = True

        return resultset

    # ----- Array row operations.

    def get_parent_keys_from_url(self) -> dict:
        """
        Get parent key values from the url's query parameters

        Returns:
            dict: The parent key values from the URL parameters.
        """
        parent_id_values = {
            v["parentElementName"]:
                self.query_params.get(v['parameterName'])
            for v in self.parent_key_names
            if self.query_params.get(v['parameterName'])
        }
        return parent_id_values

    def get_parent_key_fieldnames(self) -> list:
        """
        Get parent key names from the columns configuration.

        Returns:
            list: The parent key field names.
        """
        return [v['parameterName'] for v in self.parent_key_names]

    def get_parent_keys(self, data: dict) -> dict:
        """
        Get parent key values from given row.

        Args:
            data (dict): row with the column's values.

        Returns:
            dict: The parent key values from the row.
        """
        parent_keys = {
            v["parentElementName"]:
                data[v['parameterName']]
            for v in self.parent_key_names
        }
        if "id" in parent_keys:
            # {'_id': ObjectId(data[self.parent_key_field])}
            parent_keys["_id"] = ObjectId(parent_keys["id"])
            del parent_keys["id"]
        return parent_keys



"""
Generic Database Helper, to handle all operations over a given table.
"""
from typing import Optional, Union
from itertools import islice
import json
import re

from bson.json_util import dumps, ObjectId

from genericsuite.util.app_logger import log_debug
from genericsuite.util.datetime_utilities import (
    current_datetime_timestamp,
    get_date_range_filter,
)
from genericsuite.util.utilities import (
    get_standard_base_exception_msg,
    get_default_resultset,
    email_verification,
    sort_list_of_dicts,
)
from genericsuite.util.db_abstractor import (
    verify_required_fields,
    get_order_direction,
)
from genericsuite.util.nav_helpers import (
    put_total_pages_in_resultset,
    put_total_pages_from_resultset
)
from genericsuite.util.passwords import Passwords
from genericsuite.util.generic_db_helpers_super import GenericDbHelperSuper


DEBUG = False


class GenericDbHelper(GenericDbHelperSuper):
    """
    Generic Database Helper, to handle all operations over a given table.
    """
    def fetch_list(
        self,
        skip: int,
        limit: int,
        like_query_params: Optional[dict] = None,
        combinator: Optional[str] = None,
        order_param: Optional[Union[str, None]] = None,
    ) -> dict:
        """
        Fetches a list of items from the table based on
        the provided skip, limit, SQL-like filtering params and
        query parameters.

        Args:
            skip (int): The number of items to skip.
            limit (int): The maximum number of items to return.
                Zero means no limit.
            like_query_params (dict): The elements and its values to filter
                the items as SQL LIKE.
            combinator (str): condition combinator for the filter.
                It could be $and, $or, $not. Defaults to $and.

        Returns:
            dict: The resultset containing the list of items.
            Mandatory filters will be applied.
        """
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        if not like_query_params:
            like_query_params = {}
        if not combinator:
            combinator = '$and'

        # The resulting will be something like:
        # listing_filter BEFORE: {'meal_date': {'$lte': 946702800.0,
        # '$gte': 946616400.0}, 'observations': {'$regex':
        # '.*sancocho.*', '$options': 'si'}, 'user_id': 'XXXX'}
        listing_filter = {
            # SQL LIKE in MongoDb
            k: {
                '$regex': f".*{v}.*",
                '$options': 'si',
            } if k not in self.mandatory_filters
            and not self.is_datefield(k)
            # Date range
            else get_date_range_filter(v)
            if k not in self.mandatory_filters
            and self.is_datefield(k)
            # Normal filter
            else v
            for k, v in like_query_params.items()
            # Exclude paging and search configuration parameters
            if k not in ['page', 'limit', 'like', 'comb', 'order']
        }

        if '_id' in listing_filter:
            try:
                listing_filter['_id'] = ObjectId(listing_filter['_id'])
            except ValueError:
                resultset['error_message'] = \
                    f"_id `{listing_filter['_id']}` is invalid [FUL3]."
            except BaseException as err:
                resultset['error_message'] = \
                    get_standard_base_exception_msg(err, 'FUL4')

        if resultset['error_message']:
            resultset['error'] = True
            return resultset

        if self.query_params.get("only_listing_cols", "1") == "1":
            # By default, include only listing enabled columns and
            # unprotected columns (those not in projection_exclusions)
            projection = self.listing_disabled_columns_projection()
        else:
            # include only unprotected columns (those not in
            # projection_exclusions)
            projection = self.listing_projection_exclusions()

        # The resulting will be something like:
        # {'$or': [{'$and': [{'user_id': 'XXXX'},
        # {'meal_date': {'$lte': 946702800.0, '$gte': 946616400.0}}]},
        # {'$and': [{'user_id': 'XXXX'}, {'observations':
        # {'$regex': '.*sanncocho.*', '$options': 'si'}}]}]}
        listing_filter = self.add_mandatory_filters(listing_filter, combinator)

        _ = DEBUG and \
            log_debug(f"FETCH_LIST | self.table_name: {self.table_name}" +
                      f"\n | combinator: {combinator}" +
                      f"\n | like_query_params: {like_query_params}" +
                      f"\n | listing_filter: {listing_filter}")

        column_name, direction = self.get_sort_config(order_param)
        _ = DEBUG and \
            log_debug(f"FETCH_LIST 010 | column_name: {column_name}," +
                      f" direction: {direction}")

        try:
            db_result = (
                self.table_obj.find(
                    listing_filter, projection
                )
                .sort(
                    column_name,
                    get_order_direction(direction)
                )
            )
            if skip > 0:
                db_result = db_result.skip(int(skip))
            if limit > 0:
                db_result = db_result.limit(int(limit))
            resultset['resultset'] = dumps(db_result)
            _ = DEBUG and \
                log_debug(f"FETCH_LIST 020 | resultset: {resultset}")
        except BaseException as err:
            resultset['error_message'] = get_standard_base_exception_msg(
                err, 'FUL1'
            )
            resultset['error'] = True
        resultset = put_total_pages_in_resultset(
            self.table_obj,
            listing_filter,
            limit,
            resultset,
            'FUL2'
        )
        _ = DEBUG and \
            log_debug(f"FETCH_LIST 030 | resultset: {resultset}")
        return self.run_specific_func('list', resultset)

    def fetch_row(self, row_id: str, projection: dict = None) -> dict:
        """
        Fetches a row from the database based on the given row_id.

        Args:
            row_id (str): The ID of the row to fetch.
            projection (dict, optional): The projection to apply to the
            fetched row.

        Returns:
            dict: The resultset containing the fetched row.
        """
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        if not projection:
            projection = {}

        db_row = self.fetch_row_raw(row_id, projection)
        if not db_row['resultset']:
            resultset['error_message'] = f"Id {row_id} doesn't exist [FU1]."
        elif db_row['error']:
            resultset['error_message'] = db_row['error_message']

        if resultset['error_message']:
            resultset['error'] = True
            return resultset

        try:
            resultset['resultset'] = dumps(db_row['resultset'])
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'FU2')
            resultset['error'] = True

        return self.run_specific_func('read', resultset)

    def fetch_row_by_entryname_raw(
        self,
        entry_name: str,
        entry_value: str,
        filters: dict = None,
    ) -> dict:
        """
        Fetches a row from the database based on the given
        entry_name and entry_value and returns it without
        applying dumps() to the 'resultset' element.

        Args:
            entry_name (str): The name of the entry to filter by.
            entry_value (str): The value of the entry to filter by.
            filters (dict, optional): Additional filters to apply.
            e.g. user_id.

        Returns:
            dict: The resultset containing the fetched row.
        """
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        filters = {} if not filters else filters
        filters.update({entry_name: entry_value})
        try:
            resultset['resultset'] = self.table_obj.find_one(
                filters
            )
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'FUBEN1')
            resultset['error'] = True
        _ = DEBUG and \
            log_debug("fetch_row_by_entryname_raw: " +
                      f"entry_name: {entry_name}" +
                      f" | entry_value: {entry_value}" +
                      f" | resultset: {resultset}")
        return resultset

    def create_row(
        self,
        data: dict,
        filters: dict = None,
    ) -> dict:
        """
        Creates a new row in the database based on the given data.

        Args:
            data (dict): The data to be inserted into the database.
            filters (dict, optional): Additional filters to apply in
            pre-create verification.

        Returns:
            dict: The resultset containing the ID of the newly
            created row and the number of affected rows.
        """
        psw_class = Passwords()
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        filters = {} if not filters else filters
        data['creation_date'] = data['update_date'] = \
            current_datetime_timestamp()
        # Verify required fields are in the item data supplied
        mandatory_fields = self.get_mandatory_fields(data, is_create=True)
        resultset = verify_required_fields(data, mandatory_fields, '[CU6]')
        if resultset['error_message']:
            resultset['error'] = True
            return resultset
        # Verify emails to be valid for 'email_verification' attribute list
        if self.cnf_db.get('email_verification'):
            resultset = email_verification(
                data,
                self.cnf_db.get('email_verification')
            )
        if resultset['error_message']:
            resultset['error'] = True
            return resultset
        # Add table's mandatoy filters. E.g. user_id
        filters.update({
            k: v for k, v in self.mandatory_filters.items()
        })
        # Set secondary key attribute name and value to verify exisitng item
        pk_name = self.cnf_db.get('creation_pk_name', '_id')
        pk_value = data.get(pk_name, None)
        _ = DEBUG and \
            log_debug(f'>>>---> Create {self.table_name} | ' +
                      f'pk_name: {pk_name} | pk_value: {str(pk_value)}')
        if pk_name == '_id' and pk_value:
            pk_value = ObjectId(pk_value)
        # Verify if the item already exists
        db_row = self.fetch_row_by_entryname_raw(pk_name, pk_value, filters)
        if db_row['resultset']:
            _ = DEBUG and \
                log_debug(f'Create {self.name} | db_row["resultset"]:' +
                          f' {db_row["resultset"]}')
            resultset['error_message'] = \
                f"{self.name} {str(pk_value)} already exists [CU4]."
        elif db_row['error']:
            resultset['error_message'] = db_row['error_message']

        if resultset['error_message']:
            resultset['error'] = True
            return resultset

        # Encrypt passwords
        if self.cnf_db.get('passwords'):
            data = psw_class.passwords_encryption(
                data,
                self.cnf_db['passwords']
            )
        # Creates the new item
        try:
            resultset['resultset']['_id'] = str(
                self.table_obj.insert_one(data).inserted_id
            )
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'CU5')
            resultset['error'] = True
        else:
            resultset['resultset']['rows_affected'] = '1'

        return self.run_specific_func('create', resultset)

    def update_row(self, record: dict, options: dict = None) -> dict:
        """
        Updates a existing row in the database based on the given data.
        If password type fields are configured, it preserves it
        from the original row in case isn't included in the
        modified row (record).
        It figures out the corresponding primary key looking for
        "id" or "_id" in the record columns.

        Args:
            record (dict): The data to be inserted into the database.
            options (dict): options. Defaults to {}.
                "update_item" (str): "1" apply update_one() instead
                    of replace_one(). Defaults to "0"

        Returns:
            dict: The resultset containing the number of affected rows.
        """
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        options = {} if options is None else options
        if self.cnf_db.get('updateItem'):
            options["update_item"] = self.cnf_db.get('updateItem')
        mandatory_fields = self.get_mandatory_fields(record)
        updated_record = dict(record)
        updated_record['update_date'] = current_datetime_timestamp()

        _ = DEBUG and \
            log_debug(f'>>>---> Update {self.table_name} | ' +
                      f'updated_record (before): {updated_record}')

        resultset = verify_required_fields(updated_record,
                                           mandatory_fields, '[UU1]')
        if resultset['error']:
            return resultset

        resultset = get_default_resultset()

        if '_id' not in record and 'id' in record:
            # Only one ID should be in the original record, and must be _id,
            # because it's the field used to update the specific ID.
            record['_id'] = record['id']
            del record['id']

        if '_id' in updated_record:
            # To avoid "WriteError('Performing an update on the path '_id'
            # would modify the immutable field '_id'
            del updated_record['_id']
        if 'id' in updated_record:
            del updated_record['id']

        if self.cnf_db.get('passwords'):
            encrypt_pasw_rs = self.passwords_encryption_on_update(
                record, updated_record, self.cnf_db['passwords']
            )
            if encrypt_pasw_rs['error']:
                resultset['error_message'] = encrypt_pasw_rs['error_message']
                resultset['error'] = True
                return resultset
            updated_record = encrypt_pasw_rs['updated_record']

        _ = DEBUG and \
            log_debug(f'>>>---> PERFORM Update {self.table_name}' +
                      f' | ID: {record["_id"]}')

        try:
            if options.get("update_item", "0") == "0":
                # Replace what is was in the item
                op_result = self.table_obj.replace_one(
                    {'_id': ObjectId(record['_id'])},
                    updated_record
                ).modified_count
            else:
                # Update the item, preserving the attributes not present in
                # the updated_record
                op_result = self.table_obj.update_one(
                    {'_id': ObjectId(record['_id'])},
                    {'$set': updated_record}
                ).modified_count
            resultset['resultset']['rows_affected'] = str(op_result)
            # Ensure the _id is returned for the specific function
            resultset['resultset']['_id'] = str(record['_id'])
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'UU2')
            resultset['error'] = True

        _ = DEBUG and \
            log_debug(f'>>>---> Update {self.table_name} | ' +
                      f'updated_record (after): {updated_record}')

        return self.run_specific_func('update', resultset)

    def delete_row(self, remove_id: str) -> dict:
        """
        Deletes a existing row in the database.

        Args:
            remove_id (str): id of the row to be deleted.

        Returns:
            dict: The resultset containing the number of affected rows.
        """
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        db_row = self.fetch_row_by_entryname_raw('_id', ObjectId(remove_id))
        if not db_row['resultset']:
            resultset['error_message'] = \
                f"error: {self.name} {remove_id} doesn't exist [DU1]."
        elif db_row['error']:
            resultset['error_message'] = db_row['error_message']

        if resultset['error_message']:
            resultset['error'] = True
            return resultset

        try:
            resultset['resultset']['rows_affected'] = str(
                self.table_obj.delete_one(
                    {'_id': ObjectId(remove_id)}
                ).deleted_count
            )
            # Ensure the _id is returned for the specific function
            resultset['resultset']['_id'] = str(remove_id)
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'DU2')
            resultset['error'] = True

        return self.run_specific_func('delete', resultset)

    # ----- Array row operations.

    def fetch_array_rows(
        self,
        row_id: str,
        filters: dict = None,
        skip: int = 0,
        limit: int = None,
        like_query_params: Optional[dict] = None,
        combinator: Optional[str] = None,
        order_param: str = None,
    ) -> dict:
        """
        Fetches a list of items from a row array based on
        the provided skip, limit, SQL-like filtering params and
        query parameters.

        Args:
            row_id (str): primary key of the row that holds the array field.
            filters (dict): The elements and its values to filter
                the items, e.g. the item ID / Primary Key. Defaults to None.
            skip (int): The number of items to skip.
            limit (int): The maximum number of items to return.
                Zero means no limit.
            like_query_params (dict): The elements and its values to filter
                the items as SQL LIKE.
            combinator (str): condition combinator for the filter.
                It could be $and, $or, $not. Defaults to $and.

        Returns:
            dict: The resultset containing the list of items
            in the row's array.
        """
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        if not like_query_params:
            like_query_params = {}
        if not combinator:
            combinator = '$and'
        projection = {
            self.array_field: 1
        }

        db_parent_row = self.fetch_row_raw(row_id, projection)
        if not db_parent_row['resultset']:
            resultset['error_message'] = \
                f'Id {row_id} doesn\'t exist [FUFT1].'
        elif db_parent_row['error']:
            resultset['error_message'] = db_parent_row['error_message']
        if resultset['error_message']:
            resultset['error'] = True
            return resultset

        response = db_parent_row['resultset'].get(self.array_field, [])

        if filters is not None:
            for filter_key in filters:
                response = list(
                    filter(
                        lambda x, key=filter_key: x.get(key) == filters[key],
                        response
                    )
                )

        if len(like_query_params):
            parent_keys = self.get_parent_key_fieldnames()
            lf = {
                # SQL LIKE in MongoDb
                k: {
                    "type": 'regex',
                    "$regex": f".*{v}.*",
                } if not self.is_datefield(k)
                # Date range
                else get_date_range_filter(v, {"type": 'range'})
                if self.is_datefield(k)
                # Normal filter
                else {
                    "value": v
                }
                for k, v in like_query_params.items()
                # Exclude paging and search configuration parameters
                if k not in ['page', 'limit', 'like', 'comb', 'order']
                and k not in parent_keys
            }
            # Create a list of filters based on lf (acronim of listing filters)
            all_filters = [
                lambda x, key=filter_key:
                re.search(lf[key]["$regex"], x.get(key, ''),
                          re.IGNORECASE) is not None
                if lf[key]["type"] == "regex"
                else (x.get(key) <= lf[key]["$lte"] and
                      x.get(key) >= lf[key]["$gte"])
                if lf[key]["type"] == "range"
                else x.get(key) == lf[key]["value"]
                for filter_key in lf
            ]
            _ = DEBUG and \
                log_debug(
                    f'\nfetch_array_rows | lf: {lf}' +
                    f'\nfetch_array_rows | all_filters: {all_filters}\n')
            # Apply filters using OR logic
            response = list(
                filter(
                    lambda x: any(f(x) for f in all_filters),
                    response
                )
            )

        try:
            resultset['resultset'] = dumps(response)
        except BaseException as err:
            resultset['error_message'] = get_standard_base_exception_msg(
                err, 'FUFT2'
            )
            resultset['error'] = True
        # All entries are needed to calculate total pages
        resultset = put_total_pages_from_resultset(
            limit,
            resultset,
        )
        # Now entries are sliced to return only the entries
        # corresponding to the requested page
        limit_for_slice = None if limit == 0 else limit
        response = json.loads(resultset['resultset'])
        response = list(islice(islice(response, skip, None), limit_for_slice))

        # Sort the result
        column_name, direction = self.get_sort_config(order_param)
        response = sort_list_of_dicts(
            data=response,
            column_name=column_name,
            direction=direction
        )

        resultset['resultset'] = dumps(response)
        return self.run_specific_func('list', resultset)

    def add_array_item_to_row(self, data: dict) -> dict:
        """
        Creates a new item in a row array based on the
        given data. It can prevent already existing items
        or allow duplicates.
        The parent row primary key is determined by the
        configuration (see get_parent_keys()) and if the
        "uuid_generator" attribute is in the 'fieldElements'
        configuration, it will generate a UUID4 for the
        key_fieldname (as defined in self.array_field_key).

        Args:
            data (dict): The data to be inserted into the row.

            e.g. the row primary key is "daily_meal_id" and the
            row array is 'meal_ingredients'. In this example
            a new item will be inserted in the array with the
            values of 'food_moment_id', 'calories_value' and
            'calories_unit'.

            {'daily_meal_id': 'XXXX',
             'meal_ingredients':
                {'food_moment_id': 'YYYY',
                 'calories_value': 110,
                 'calories_unit': 'kcal',
                }
            }

        Returns:
            dict: The resultset containing the ID of the newly
            created row and the number of affected rows.
        """
        _ = DEBUG and \
            log_debug('\nadd_array_item_to_row - data' +
                      f'{data}\n')
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        parent_keys = self.get_parent_keys(data)

        if not self.allow_duplicates:
            # Get the array field ID, if not exists, returns an error
            try:
                key_value = data[self.array_field].get(self.array_field_key)
            except BaseException as err:
                resultset['error_message'] = \
                    get_standard_base_exception_msg(err, 'AFTTU2')
                resultset['error'] = True
                return resultset
            # Verify array field ID dupes
            if key_value:
                key_existence = self.get_array_item_in_row(key_value,
                                                           parent_keys)
                if key_existence['error']:
                    return key_existence
                if int(key_existence['resultset']['rows_count']) > 0:
                    resultset['error_message'] = \
                        f'error: {self.name} ' + \
                        f'{key_value} already exist [AFTTU3].'
                    resultset['error'] = True
                    return resultset

        # Check UUID generator
        data = self.uuid_generator_check(data, self.array_field_key,
                                         self.array_field)

        # Add the array field ID and time
        add_to_set = {
            self.array_field: data[self.array_field]
        }
        try:
            resultset['resultset']['rows_affected'] = str(
                self.table_obj.update_one(
                    parent_keys,
                    {'$addToSet': add_to_set}
                ).modified_count
            )
            # Ensure the _id is returned for the specific function
            resultset['resultset']['_id'] = str(parent_keys['_id'])
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'AFTTU1')
            resultset['error'] = True
        return self.run_specific_func('create', resultset)

    def remove_array_item_from_row(self, data: dict) -> dict:
        """
        Removes one item in a row array based on the given data.
        The parent row primary key is determined by the
        configuration (see get_parent_keys())

        E.g. 'data' given is:

            {'daily_meal_id': 'XXXX',
             'meal_ingredients':
                {'food_moment_id': 'YYYY',
                 'calories_value': 110,
                 'calories_unit': 'kcal',
                 'id': 'ZZZZ',
                 'quantity': 1, ...
                },
             'meal_ingredients_old':
                {'food_moment_id': 'YYYY',
                 'calories_value': 110,
                 'calories_unit': 'kcal',
                 'id': 'ZZZZ',
                 'quantity': 2, ...
                }
            }

        The self.array_field configuration is "meal_ingredients",
        "meal_ingredients_old" exist in 'data', and
        self.array_field_key configuration is "id",

        So it will $pull (remove) the element 'id': 'ZZZZ'
        as pull_element will be:
            pull_element={'meal_ingredients': {'id': 'ZZZZ'}}
        and parent_keys (row's PK) is: {'_id': ObjectId('XXXX')}

        Args:
            data (dict): The data to be inserted into the row.

        Returns:
            dict: The resultset containing the number of affected rows.
        """
        _ = DEBUG and \
            log_debug('\nremove_array_item_from_row - data' +
                      f'{data}\n')
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        array_field_in_json = self.array_field
        if f'{array_field_in_json}_old' in data:
            # This is for older entry deletion,
            # when the key field has been changed
            array_field_in_json = f'{array_field_in_json}_old'
        _ = DEBUG and \
            log_debug(f'\n$pull from "{self.array_field}", ' +
                      f'array_field_key={self.array_field_key}' +
                      f'\narray_field_in_json={array_field_in_json}, ' +
                      'key value to REMOVE=' +
                      f'{data[array_field_in_json][self.array_field_key]}\n')
        pull_element = {
            self.array_field: {
                self.array_field_key:
                    data[array_field_in_json][self.array_field_key]
            }
        }
        _ = DEBUG and \
            log_debug(f'pull_element={pull_element}')

        parent_keys = self.get_parent_keys(data)

        _ = DEBUG and \
            log_debug(f'parent_keys: {parent_keys}\n')
        try:
            resultset['resultset']['rows_affected'] = str(
                self.table_obj.update_one(
                    parent_keys,
                    {'$pull': pull_element}
                ).modified_count
            )
            # Ensure the _id is returned for the specific function
            resultset['resultset']['_id'] = str(parent_keys['_id'])
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'RFTTU')
            resultset['error'] = True
        return self.run_specific_func('delete', resultset)

    def get_array_item_in_row(self, array_item_id: str,
                              parent_keys: dict) -> dict:
        """
        Retrieves the items count in a row's array.

        Args:
            array_item_id (str): The ID of the array item to search for.
            parent_keys (dict): The keys identifying the parent row.

        Returns:
            dict: The resultset containing the count of items in the row.
        """
        _ = DEBUG and \
            log_debug('\nget_array_item_in_row - array_item_id:' +
                      f' {str(array_item_id)}' +
                      f' | parent_keys: {str(parent_keys)}\n')
        resultset = get_default_resultset()
        if self.error_message:
            resultset['error_message'] = self.error_message
            resultset['error'] = True
            return resultset

        find_criteria = dict(parent_keys)
        find_criteria.update({
            self.array_field: {
                '$elemMatch': {
                    self.array_field_key: array_item_id
                }
            },
        })
        _ = DEBUG and \
            log_debug('self.table_obj | find_criteria:' +
                      f'{find_criteria}\n')
        try:
            resultset['resultset']['rows_count'] = str(
                self.table_obj.count_documents(find_criteria)
            )
        except BaseException as err:
            resultset['error_message'] = \
                get_standard_base_exception_msg(err, 'GFMIU-010')
            resultset['error'] = True
        _ = DEBUG and \
            log_debug(f">>--> get_array_item_in_row | resultset: {resultset}")
        return self.run_specific_func('read', resultset)
